@@ -28,6 +28,7 @@
 #include <smpp34.h>
 #include <smpp34_structs.h>
 #include <smpp34_params.h>
+#include <smpp34_heap.h>
 
 #include <osmocom/core/utils.h>
 #include <osmocom/core/msgb.h>
@@ -45,9 +46,38 @@
 #include <osmocom/msc/transaction.h>
 #include <osmocom/msc/gsm_subscriber.h>
 #include <osmocom/msc/vlr.h>
+#include <osmocom/msc/msc_a.h>
 
 #include "smpp_smsc.h"
 
+#define VSUB_USE_SMPP "SMPP"
+#define VSUB_USE_SMPP_CMD "SMPP-cmd"
+
+/* talloc integration for libsmpp34 */
+
+static struct smsc *g_smsc;
+
+static void *smpp34_talloc_malloc(size_t sz)
+{
+	return talloc_size(g_smsc, sz);
+}
+
+static void *smpp34_talloc_realloc(void *ptr, size_t sz)
+{
+	return talloc_realloc_size(g_smsc, ptr, sz);
+}
+
+static void smpp34_talloc_free(void *ptr)
+{
+	talloc_free(ptr);
+}
+
+static const struct smpp34_memory_functions smpp34_talloc = {
+	.malloc_fun = smpp34_talloc_malloc,
+	.realloc_fun = smpp34_talloc_realloc,
+	.free_fun = smpp34_talloc_free,
+};
+
 /*! \brief find vlr_subscr for a given SMPP NPI/TON/Address */
 static struct vlr_subscr *subscr_by_dst(struct gsm_network *net,
 					    uint8_t npi, uint8_t ton,
@@ -57,11 +87,11 @@ static struct vlr_subscr *subscr_by_dst(struct gsm_network *net,
 
 	switch (npi) {
 	case NPI_Land_Mobile_E212:
-		vsub = vlr_subscr_find_by_imsi(net->vlr, addr);
+		vsub = vlr_subscr_find_by_imsi(net->vlr, addr, VSUB_USE_SMPP);
 		break;
 	case NPI_ISDN_E163_E164:
 	case NPI_Private:
-		vsub = vlr_subscr_find_by_msisdn(net->vlr, addr);
+		vsub = vlr_subscr_find_by_msisdn(net->vlr, addr, VSUB_USE_SMPP);
 		break;
 	default:
 		LOGP(DSMPP, LOGL_NOTICE, "Unsupported NPI: %u\n", npi);
@@ -117,7 +147,7 @@ static int submit_to_sms(struct gsm_sms **psms, struct gsm_network *net,
 			if (smpp34_submit_tlv_msg_payload(t, submit, &sms_msg,
 							  &sms_msg_len) < 0) {
 				if (dest)
-					vlr_subscr_put(dest);
+					vlr_subscr_put(dest, VSUB_USE_SMPP);
 				return ESME_ROPTPARNOTALLWD;
 			}
 			break;
@@ -136,7 +166,8 @@ static int submit_to_sms(struct gsm_sms **psms, struct gsm_network *net,
 		} else {
 			LOGP(DLSMS, LOGL_ERROR,
 			     "SMPP neither message payload nor valid sm_length.\n");
-			vlr_subscr_put(dest);
+			if (dest)
+				vlr_subscr_put(dest, VSUB_USE_SMPP);
 			return ESME_RINVPARLEN;
 		}
 	}
@@ -149,6 +180,12 @@ static int submit_to_sms(struct gsm_sms **psms, struct gsm_network *net,
 
 	/* fill in the destination address */
 	sms->receiver = dest;
+	if (dest) {
+		/* Replace use count from above subscr_by_dst (VSUB_USE_SMPP) by the sms->receiver use count
+		 * (VSUB_USE_SMS_RECEIVER) */
+		vlr_subscr_get(sms->receiver, VSUB_USE_SMS_RECEIVER);
+		vlr_subscr_put(dest, VSUB_USE_SMPP);
+	}
 	sms->dst.ton = submit->dest_addr_ton;
 	sms->dst.npi = submit->dest_addr_npi;
 	if (dest)
@@ -268,7 +305,7 @@ int handle_smpp_submit(struct osmo_esme *esme, struct submit_sm_t *submit,
 		LOGP(DLSMS, LOGL_DEBUG, "SMPP SUBMIT-SM: Forwarding in "
 			"real time (Transaction/Forward mode)\n");
 		sms->smpp.transaction_mode = 1;
-		gsm411_send_sms_subscr(sms->receiver, sms);
+		gsm411_send_sms(net, sms->receiver, sms);
 		rc = 1; /* don't send any response yet */
 		break;
 	}
@@ -282,9 +319,19 @@ static void alert_all_esme(struct smsc *smsc, struct vlr_subscr *vsub,
 
 	llist_for_each_entry(esme, &smsc->esme_list, list) {
 		/* we currently send an alert notification to each ESME that is
-		 * connected, and do not require a (non-existant) delivery
-		 * pending flag to be set before,  FIXME: make this VTY
-		 * configurable */
+		 * connected, and do not require a (non-existent) delivery
+		 * pending flag to be set before. */
+		if (!esme->bind_flags) {
+			LOGP(DSMPP, LOGL_DEBUG,
+				"ESME is not (yet) bound, skipping alert\n");
+			continue;
+		}
+		if (esme->acl && !esme->acl->alert_notifications) {
+			LOGP(DSMPP, LOGL_DEBUG,
+				"[%s] is not set to receive Alert Notifications\n",
+					esme->system_id);
+			continue;
+		}
 		if (esme->acl && esme->acl->deliver_src_imsi) {
 			smpp_tx_alert(esme, TON_Subscriber_Number,
 				      NPI_Land_Mobile_E212,
@@ -320,7 +367,7 @@ static int smpp_sms_cb(unsigned int subsys, unsigned int signal,
 		 * to the ESME */
 	case S_SMS_UNKNOWN_ERROR:
 		if (sms->smpp.transaction_mode) {
-			/* Send back the SUBMIT-SM response with apropriate error */
+			/* Send back the SUBMIT-SM response with appropriate error */
 			LOGP(DLSMS, LOGL_INFO, "SMPP SUBMIT-SM: Error\n");
 			rc = smpp_tx_submit_r(sms->smpp.esme,
 					      sms->smpp.sequence_nr,
@@ -522,7 +569,7 @@ static void smpp_cmd_free(struct osmo_smpp_cmd *cmd)
 {
 	osmo_timer_del(&cmd->response_timer);
 	llist_del(&cmd->list);
-	vlr_subscr_put(cmd->vsub);
+	vlr_subscr_put(cmd->vsub, VSUB_USE_SMPP_CMD);
 	talloc_free(cmd);
 }
 
@@ -536,22 +583,21 @@ void smpp_cmd_flush_pending(struct osmo_esme *esme)
 
 void smpp_cmd_ack(struct osmo_smpp_cmd *cmd)
 {
-	struct gsm_subscriber_connection *conn;
+	struct msc_a *msc_a;
 	struct gsm_trans *trans;
 
 	if (cmd->is_report)
 		goto out;
 
-	conn = connection_for_subscr(cmd->vsub);
-	if (!conn) {
-		LOGP(DSMPP, LOGL_ERROR, "No connection to subscriber anymore\n");
+	msc_a = msc_a_for_vsub(cmd->vsub, true);
+	if (!msc_a) {
+		LOGP(DSMPP, LOGL_ERROR, "No connection to subscriber %s\n", vlr_subscr_name(cmd->vsub));
 		goto out;
 	}
 
-	trans = trans_find_by_id(conn, GSM48_PDISC_SMS, cmd->gsm411_trans_id);
+	trans = trans_find_by_id(msc_a, TRANS_SMS, cmd->gsm411_trans_id);
 	if (!trans) {
-		LOGP(DSMPP, LOGL_ERROR, "GSM transaction %u is gone\n",
-		     cmd->gsm411_trans_id);
+		LOG_MSC_A_CAT(msc_a, DSMPP, LOGL_ERROR, "GSM transaction %u is gone\n", cmd->gsm411_trans_id);
 		goto out;
 	}
 
@@ -562,23 +608,23 @@ out:
 
 void smpp_cmd_err(struct osmo_smpp_cmd *cmd, uint32_t status)
 {
-	struct gsm_subscriber_connection *conn;
+	struct msc_a *msc_a;
 	struct gsm_trans *trans;
 	int gsm411_cause;
 
 	if (cmd->is_report)
 		goto out;
 
-	conn = connection_for_subscr(cmd->vsub);
-	if (!conn) {
-		LOGP(DSMPP, LOGL_ERROR, "No connection to subscriber anymore\n");
+	msc_a = msc_a_for_vsub(cmd->vsub, true);
+	if (!msc_a) {
+		LOGP(DSMPP, LOGL_ERROR, "No connection to subscriber %s\n", vlr_subscr_name(cmd->vsub));
 		goto out;
 	}
 
-	trans = trans_find_by_id(conn, GSM48_PDISC_SMS, cmd->gsm411_trans_id);
+	trans = trans_find_by_id(msc_a, TRANS_SMS, cmd->gsm411_trans_id);
 	if (!trans) {
-		LOGP(DSMPP, LOGL_ERROR, "GSM transaction %u is gone\n",
-		     cmd->gsm411_trans_id);
+		LOG_MSC_A_CAT(msc_a, DSMPP, LOGL_ERROR, "GSM transaction %u is gone\n",
+			      cmd->gsm411_trans_id);
 		goto out;
 	}
 
@@ -609,7 +655,8 @@ static int smpp_cmd_enqueue(struct osmo_esme *esme,
 	cmd->is_report		= sms->is_report;
 	cmd->gsm411_msg_ref	= sms->gsm411.msg_ref;
 	cmd->gsm411_trans_id	= sms->gsm411.transaction_id;
-	cmd->vsub		= vlr_subscr_get(vsub);
+	vlr_subscr_get(vsub, VSUB_USE_SMPP_CMD);
+	cmd->vsub = vsub;
 
 	/* FIXME: No predefined value for this response_timer as specified by
 	 * SMPP 3.4 specs, section 7.2. Make this configurable? Don't forget
@@ -636,11 +683,12 @@ struct osmo_smpp_cmd *smpp_cmd_find_by_seqnum(struct osmo_esme *esme,
 }
 
 static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
-			   struct gsm_subscriber_connection *conn)
+			   struct msc_a *msc_a)
 {
 	struct deliver_sm_t deliver;
 	int mode, ret;
 	uint8_t dcs;
+	struct vlr_subscr *vsub = msc_a_vsub(msc_a);
 
 	memset(&deliver, 0, sizeof(deliver));
 	deliver.command_length	= 0;
@@ -653,13 +701,13 @@ static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
 		deliver.source_addr_npi = NPI_Land_Mobile_E212;
 		snprintf((char *)deliver.source_addr,
 			sizeof(deliver.source_addr), "%s",
-			conn->vsub->imsi);
+			vsub->imsi);
 	} else {
 		deliver.source_addr_ton = TON_Network_Specific;
 		deliver.source_addr_npi = NPI_ISDN_E163_E164;
 		snprintf((char *)deliver.source_addr,
 			 sizeof(deliver.source_addr), "%s",
-			 conn->vsub->msisdn);
+			 vsub->msisdn);
 	}
 
 	deliver.dest_addr_ton	= sms->dst.ton;
@@ -725,19 +773,18 @@ static int deliver_to_esme(struct osmo_esme *esme, struct gsm_sms *sms,
 	if (ret < 0)
 		return ret;
 
-	return smpp_cmd_enqueue(esme, conn->vsub, sms,
+	return smpp_cmd_enqueue(esme, vsub, sms,
 				deliver.sequence_number);
 }
 
 static struct smsc *g_smsc;
 
-int smpp_route_smpp_first(struct gsm_sms *sms, struct gsm_subscriber_connection *conn)
+bool smpp_route_smpp_first()
 {
-	return g_smsc->smpp_first;
+	return (bool)(g_smsc->smpp_first);
 }
 
-int smpp_try_deliver(struct gsm_sms *sms,
-		     struct gsm_subscriber_connection *conn)
+int smpp_try_deliver(struct gsm_sms *sms, struct msc_a *msc_a)
 {
 	struct osmo_esme *esme;
 	struct osmo_smpp_addr dst;
@@ -750,7 +797,7 @@ int smpp_try_deliver(struct gsm_sms *sms,
 
 	rc = smpp_route(g_smsc, &dst, &esme);
 	if (!rc)
-		rc = deliver_to_esme(esme, sms, conn);
+		rc = deliver_to_esme(esme, sms, msc_a);
 
 	return rc;
 }
@@ -769,6 +816,7 @@ int smpp_openbsc_alloc_init(void *ctx)
 		LOGP(DSMPP, LOGL_FATAL, "Cannot allocate smsc struct\n");
 		return -1;
 	}
+	smpp34_set_memory_functions(&smpp34_talloc);
 	return smpp_vty_init();
 }
 
