@@ -1,4 +1,4 @@
-/* SMS queue to continously attempt to deliver SMS */
+/* SMS queue to continuously attempt to deliver SMS */
 /*
  * (C) 2010 by Holger Hans Peter Freyther <zecke@selfish.org>
  * All Rights Reserved
@@ -50,6 +50,7 @@ struct gsm_sms_pending {
 	struct llist_head entry;
 
 	struct vlr_subscr *vsub;
+	struct msc_a *msc_a;
 	unsigned long long sms_id;
 	int failed_attempts;
 	int resend;
@@ -65,7 +66,7 @@ struct gsm_sms_queue {
 
 	struct llist_head pending_sms;
 
-	char last_msisdn[GSM_EXTENSION_LENGTH+1];
+	char last_msisdn[GSM23003_MSISDN_MAX_DIGITS+1];
 };
 
 static int sms_subscr_cb(unsigned int, unsigned int, void *, void *);
@@ -118,14 +119,15 @@ static struct gsm_sms_pending *sms_pending_from(struct gsm_sms_queue *smsq,
 	if (!pending)
 		return NULL;
 
-	pending->vsub = vlr_subscr_get(sms->receiver);
+	vlr_subscr_get(sms->receiver, VSUB_USE_SMS_PENDING);
+	pending->vsub = sms->receiver;
 	pending->sms_id = sms->id;
 	return pending;
 }
 
 static void sms_pending_free(struct gsm_sms_pending *pending)
 {
-	vlr_subscr_put(pending->vsub);
+	vlr_subscr_put(pending->vsub, VSUB_USE_SMS_PENDING);
 	llist_del(&pending->entry);
 	talloc_free(pending);
 }
@@ -151,16 +153,16 @@ static void sms_pending_failed(struct gsm_sms_pending *pending, int paging_error
 	struct gsm_network *net = pending->vsub->vlr->user_ctx;
 	struct gsm_sms_queue *smsq;
 
+	pending->failed_attempts++;
 	LOGP(DLSMS, LOGL_NOTICE, "Sending SMS %llu failed %d times.\n",
 	     pending->sms_id, pending->failed_attempts);
 
 	smsq = net->sms_queue;
-	if (++pending->failed_attempts < smsq->max_fail)
+	if (pending->failed_attempts < smsq->max_fail)
 		return sms_pending_resend(pending);
 
 	sms_pending_free(pending);
 	smsq->pending -= 1;
-	sms_queue_trigger(smsq);
 }
 
 /*
@@ -186,7 +188,7 @@ static void sms_resend_pending(void *_data)
 			sms_queue_trigger(smsq);
 		} else {
 			pending->resend = 0;
-			gsm411_send_sms_subscr(sms->receiver, sms);
+			gsm411_send_sms(smsq->network, sms->receiver, sms);
 		}
 	}
 }
@@ -225,8 +227,15 @@ struct gsm_sms *smsq_take_next_sms(struct gsm_network *net,
 		osmo_strlcpy(last_msisdn, sms->dst.addr, last_msisdn_buflen);
 
 		/* Is the subscriber attached? If not, go to next SMS */
-		if (!sms->receiver || !sms->receiver->lu_complete)
+		if (!sms->receiver || !sms->receiver->lu_complete) {
+			LOGP(DLSMS, LOGL_DEBUG,
+			     "Subscriber %s%s is not attached, skipping SMS %llu\n",
+			     sms->receiver ? "" : "MSISDN-",
+			     sms->receiver ? vlr_subscr_msisdn_or_name(sms->receiver)
+					   : sms->dst.addr, sms->id);
+			sms_free(sms);
 			continue;
+		}
 
 		return sms;
 	}
@@ -247,7 +256,7 @@ static void sms_submit_pending(void *_data)
 	unsigned long long first_sub = 0;
 	int attempted = 0, rounds = 0;
 
-	LOGP(DLSMS, LOGL_DEBUG, "Attempting to send %d SMS\n", attempts);
+	LOGP(DLSMS, LOGL_DEBUG, "Attempting to send up to %d SMS\n", attempts);
 
 	do {
 		struct gsm_sms_pending *pending;
@@ -263,7 +272,7 @@ static void sms_submit_pending(void *_data)
 		}
 
 		rounds += 1;
-		LOGP(DLSMS, LOGL_DEBUG, "Sending SMS round %d\n", rounds);
+		LOGP(DLSMS, LOGL_DEBUG, "Checking whether to send SMS %llu\n", sms->id);
 
 		/*
 		 * This code needs to detect a loop. It assumes that no SMS
@@ -311,7 +320,7 @@ static void sms_submit_pending(void *_data)
 		attempted += 1;
 		smsq->pending += 1;
 		llist_add_tail(&pending->entry, &smsq->pending_sms);
-		gsm411_send_sms_subscr(sms->receiver, sms);
+		gsm411_send_sms(smsq->network, sms->receiver, sms);
 	} while (attempted < attempts && rounds < 1000);
 
 	LOGP(DLSMS, LOGL_DEBUG, "SMSqueue added %d messages in %d rounds\n", attempted, rounds);
@@ -349,7 +358,7 @@ static void sms_send_next(struct vlr_subscr *vsub)
 
 	smsq->pending += 1;
 	llist_add_tail(&pending->entry, &smsq->pending_sms);
-	gsm411_send_sms_subscr(sms->receiver, sms);
+	gsm411_send_sms(smsq->network, sms->receiver, sms);
 	return;
 
 no_pending_sms:
@@ -398,7 +407,6 @@ static int sub_ready_for_sm(struct gsm_network *net, struct vlr_subscr *vsub)
 {
 	struct gsm_sms *sms;
 	struct gsm_sms_pending *pending;
-	struct gsm_subscriber_connection *conn;
 
 	/*
 	 * The code used to be very clever and tried to submit
@@ -423,15 +431,12 @@ static int sub_ready_for_sm(struct gsm_network *net, struct vlr_subscr *vsub)
 		return 0;
 	}
 
-	conn = connection_for_subscr(vsub);
-	if (!conn)
-		return -1;
-
 	/* Now try to deliver any pending SMS to this sub */
 	sms = db_sms_get_unsent_for_subscr(vsub, UINT_MAX);
 	if (!sms)
 		return -1;
-	gsm411_send_sms(conn, sms);
+
+	gsm411_send_sms(net, vsub, sms);
 	return 0;
 }
 
@@ -479,12 +484,13 @@ static int sms_sms_cb(unsigned int subsys, unsigned int signal,
 	case S_SMS_DELIVERED:
 		/* Remember the subscriber and clear the pending entry */
 		network->sms_queue->pending -= 1;
-		vsub = vlr_subscr_get(pending->vsub);
+		vsub = pending->vsub;
+		vlr_subscr_get(vsub, __func__);
 		db_sms_delete_sent_message_by_id(pending->sms_id);
 		sms_pending_free(pending);
 		/* Attempt to send another SMS to this subscriber */
 		sms_send_next(vsub);
-		vlr_subscr_put(vsub);
+		vlr_subscr_put(vsub, __func__);
 		break;
 	case S_SMS_MEM_EXCEEDED:
 		network->sms_queue->pending -= 1;
@@ -502,23 +508,12 @@ static int sms_sms_cb(unsigned int subsys, unsigned int signal,
 		 * subscriber. If we have some kind of other transmit error we
 		 * should flag the SMS as bad.
 		 */
-		switch (sig_sms->paging_result) {
-		case 0:
+		if (sig_sms->paging_result) {
 			/* BAD SMS? */
 			db_sms_inc_deliver_attempts(sig_sms->sms);
 			sms_pending_failed(pending, 0);
-			break;
-		case GSM_PAGING_EXPIRED:
+		} else {
 			sms_pending_failed(pending, 1);
-			break;
-		case GSM_PAGING_BUSY:
-			network->sms_queue->pending -= 1;
-			sms_pending_free(pending);
-			sms_queue_trigger(network->sms_queue);
-			break;
-		default:
-			LOGP(DLSMS, LOGL_ERROR, "Unhandled result: %d\n",
-			     sig_sms->paging_result);
 		}
 		break;
 	default:
