@@ -24,30 +24,47 @@
 #include <osmocom/msc/debug.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/msc/gsm_04_08.h>
-#include <osmocom/msc/mncc.h>
-#include <osmocom/msc/osmo_msc.h>
 #include <osmocom/msc/vlr.h>
+#include <osmocom/msc/msc_a.h>
+#include <osmocom/msc/msub.h>
+#include <osmocom/msc/paging.h>
+#include <osmocom/msc/silent_call.h>
 
 void *tall_trans_ctx;
 
 void _gsm48_cc_trans_free(struct gsm_trans *trans);
+void _gsm411_sms_trans_free(struct gsm_trans *trans);
+void _gsm911_nc_ss_trans_free(struct gsm_trans *trans);
+
+struct gsm_trans *trans_find_by_type(const struct msc_a *msc_a, enum trans_type type)
+{
+	struct gsm_trans *trans;
+	struct gsm_network *net = msc_a_net(msc_a);
+	struct vlr_subscr *vsub = msc_a_vsub(msc_a);
+
+	llist_for_each_entry(trans, &net->trans_list, entry) {
+		if (trans->vsub == vsub && trans->type == type)
+			return trans;
+	}
+	return NULL;
+}
 
 /*! Find a transaction in connection for given protocol + transaction ID
- * \param[in] conn Connection in whihc we want to find transaction
+ * \param[in] conn Connection in which we want to find transaction
  * \param[in] proto Protocol of transaction
  * \param[in] trans_id Transaction ID of transaction
  * \returns Matching transaction, if any
  */
-struct gsm_trans *trans_find_by_id(struct gsm_subscriber_connection *conn,
-				   uint8_t proto, uint8_t trans_id)
+struct gsm_trans *trans_find_by_id(const struct msc_a *msc_a,
+				   enum trans_type type, uint8_t trans_id)
 {
 	struct gsm_trans *trans;
-	struct gsm_network *net = conn->network;
-	struct vlr_subscr *vsub = conn->vsub;
+	struct gsm_network *net = msc_a_net(msc_a);
+	struct vlr_subscr *vsub = msc_a_vsub(msc_a);
 
 	llist_for_each_entry(trans, &net->trans_list, entry) {
 		if (trans->vsub == vsub &&
-		    trans->protocol == proto &&
+		    trans->type == type &&
 		    trans->transaction_id == trans_id)
 			return trans;
 	}
@@ -59,7 +76,7 @@ struct gsm_trans *trans_find_by_id(struct gsm_subscriber_connection *conn,
  * \param[in] callref Call Reference of transaction
  * \returns Matching transaction, if any
  */
-struct gsm_trans *trans_find_by_callref(struct gsm_network *net,
+struct gsm_trans *trans_find_by_callref(const struct gsm_network *net,
 					uint32_t callref)
 {
 	struct gsm_trans *trans;
@@ -71,6 +88,33 @@ struct gsm_trans *trans_find_by_callref(struct gsm_network *net,
 	return NULL;
 }
 
+/*! Find a transaction by SM-RP-MR (RP Message Reference)
+ * \param[in] net Network in which we should search
+ * \param[in] vsub Subscriber for which we should search
+ * \param[in] sm_rp_mr RP Message Reference (see GSM TS 04.11, section 8.2.3)
+ * \returns Matching transaction, NULL otherwise
+ */
+struct gsm_trans *trans_find_by_sm_rp_mr(const struct gsm_network *net,
+					 const struct vlr_subscr *vsub,
+					 uint8_t sm_rp_mr)
+{
+	struct gsm_trans *trans;
+
+	llist_for_each_entry(trans, &net->trans_list, entry) {
+		if (trans->vsub == vsub &&
+		    trans->type == TRANS_SMS &&
+		    trans->sms.sm_rp_mr == sm_rp_mr)
+			return trans;
+	}
+
+	return NULL;
+}
+
+static const char *trans_vsub_use(enum trans_type type)
+{
+	return get_value_string_or_null(trans_type_names, type) ? : "trans-type-unknown";
+}
+
 /*! Allocate a new transaction and add it to network list
  *  \param[in] net Netwokr in which we allocate transaction
  *  \param[in] subscr Subscriber for which we allocate transaction
@@ -80,34 +124,38 @@ struct gsm_trans *trans_find_by_callref(struct gsm_network *net,
  */
 struct gsm_trans *trans_alloc(struct gsm_network *net,
 			      struct vlr_subscr *vsub,
-			      uint8_t protocol, uint8_t trans_id,
+			      enum trans_type type, uint8_t trans_id,
 			      uint32_t callref)
 {
+	int subsys = trans_log_subsys(type);
 	struct gsm_trans *trans;
 
 	/* a valid subscriber is indispensable */
 	if (vsub == NULL) {
-		LOGP(DCC, LOGL_NOTICE,
-		     "unable to alloc transaction, invalid subscriber (NULL)\n");
+		LOGP(subsys, LOGL_ERROR, "unable to alloc transaction, invalid subscriber (NULL)\n");
 		return NULL;
 	}
 
-	DEBUGP(DCC, "(ti %02x sub %s callref %x) New transaction\n",
-	       trans_id, vlr_subscr_name(vsub), callref);
-
-	trans = talloc_zero(tall_trans_ctx, struct gsm_trans);
+	trans = talloc(tall_trans_ctx, struct gsm_trans);
 	if (!trans)
 		return NULL;
 
-	trans->vsub = vlr_subscr_get(vsub);
-
-	trans->protocol = protocol;
-	trans->transaction_id = trans_id;
-	trans->callref = callref;
-
-	trans->net = net;
+	*trans = (struct gsm_trans){
+		.vsub = vsub,
+		.type = type,
+		.log_subsys = subsys,
+		.transaction_id = trans_id,
+		.callref = callref,
+		.net = net,
+		/* empty bearer_cap: make sure the speech_ver array is empty */
+		.bearer_cap = {
+			.speech_ver = { -1 },
+		},
+	};
+	vlr_subscr_get(vsub, trans_vsub_use(type));
 	llist_add_tail(&trans->entry, &net->trans_list);
 
+	LOG_TRANS(trans, LOGL_DEBUG, "New transaction\n");
 	return trans;
 }
 
@@ -116,71 +164,83 @@ struct gsm_trans *trans_alloc(struct gsm_network *net,
  */
 void trans_free(struct gsm_trans *trans)
 {
-	enum msc_subscr_conn_use conn_usage_token = MSC_CONN_USE_UNTRACKED;
-	struct gsm_subscriber_connection *conn;
+	const char *usage_token;
+	struct msc_a *msc_a;
 
-	switch (trans->protocol) {
-	case GSM48_PDISC_CC:
+	LOG_TRANS(trans, LOGL_DEBUG, "Freeing transaction\n");
+
+	switch (trans->type) {
+	case TRANS_CC:
 		_gsm48_cc_trans_free(trans);
-		conn_usage_token = MSC_CONN_USE_TRANS_CC;
+		usage_token = MSC_A_USE_CC;
 		break;
-	case GSM48_PDISC_SMS:
+	case TRANS_SMS:
 		_gsm411_sms_trans_free(trans);
-		conn_usage_token = MSC_CONN_USE_TRANS_SMS;
+		usage_token = MSC_A_USE_SMS;
+		break;
+	case TRANS_USSD:
+		_gsm911_nc_ss_trans_free(trans);
+		usage_token = MSC_A_USE_NC_SS;
+		break;
+	case TRANS_SILENT_CALL:
+		trans_silent_call_free(trans);
+		usage_token = MSC_A_USE_SILENT_CALL;
+		break;
+	default:
+		usage_token = NULL;
 		break;
 	}
 
 	if (trans->paging_request) {
-		subscr_remove_request(trans->paging_request);
+		paging_request_remove(trans->paging_request);
 		trans->paging_request = NULL;
 	}
 
 	if (trans->vsub) {
-		vlr_subscr_put(trans->vsub);
+		vlr_subscr_put(trans->vsub, trans_vsub_use(trans->type));
 		trans->vsub = NULL;
 	}
 
-	conn = trans->conn;
-	trans->conn = NULL;
+	msc_a = trans->msc_a;
+	trans->msc_a = NULL;
+
 	llist_del(&trans->entry);
 	talloc_free(trans);
 
-	if (conn)
-		msc_subscr_conn_put(conn, conn_usage_token);
+	if (msc_a && usage_token)
+		msc_a_put(msc_a, usage_token);
 }
 
 /*! allocate an unused transaction ID for the given subscriber
- * in the given protocol using the ti_flag specified
+ * in the given protocol using TI flag = 0 (allocated by us).
+ * See GSM 04.07, section 11.2.3.1.3 "Transaction identifier".
  * \param[in] net GSM network
- * \param[in] subscr Subscriber for which to find ID
- * \param[in] protocol Protocol for whihc to find ID
- * \param[in] ti_flag FIXME
+ * \param[in] subscr Subscriber for which to assign a new TID
+ * \param[in] protocol Protocol of to be assigned TID
  */
-int trans_assign_trans_id(struct gsm_network *net, struct vlr_subscr *vsub,
-			  uint8_t protocol, uint8_t ti_flag)
+int trans_assign_trans_id(const struct gsm_network *net, const struct vlr_subscr *vsub,
+			  enum trans_type type)
 {
 	struct gsm_trans *trans;
 	unsigned int used_tid_bitmask = 0;
 	int i, j, h;
-
-	if (ti_flag)
-		ti_flag = 0x8;
+	uint8_t proto = trans_type_to_gsm48_proto(type);
 
 	/* generate bitmask of already-used TIDs for this (subscr,proto) */
 	llist_for_each_entry(trans, &net->trans_list, entry) {
 		if (trans->vsub != vsub ||
-		    trans->protocol != protocol ||
-		    trans->transaction_id == 0xff)
+		    proto != trans_type_to_gsm48_proto(trans->type) ||
+		    trans->transaction_id == TRANS_ID_UNASSIGNED)
 			continue;
 		used_tid_bitmask |= (1 << trans->transaction_id);
 	}
 
 	/* find a new one, trying to go in a 'circular' pattern */
 	for (h = 6; h > 0; h--)
-		if (used_tid_bitmask & (1 << (h | ti_flag)))
+		if (used_tid_bitmask & (1 << h))
 			break;
 	for (i = 0; i < 7; i++) {
-		j = ((h + i) % 7) | ti_flag;
+		j = (h + i) % 7;
 		if ((used_tid_bitmask & (1 << j)) == 0)
 			return j;
 	}
@@ -190,14 +250,15 @@ int trans_assign_trans_id(struct gsm_network *net, struct vlr_subscr *vsub,
 
 /*! Check if we have any transaction for given connection
  * \param[in] conn Connection to check
- * \returns 1 in case there is a transaction, 0 otherwise
+ * \returns transaction pointer if found, NULL otherwise
  */
-struct gsm_trans *trans_has_conn(const struct gsm_subscriber_connection *conn)
+struct gsm_trans *trans_has_conn(const struct msc_a *msc_a)
 {
 	struct gsm_trans *trans;
+	struct gsm_network *net = msc_a_net(msc_a);
 
-	llist_for_each_entry(trans, &conn->network->trans_list, entry)
-		if (trans->conn == conn)
+	llist_for_each_entry(trans, &net->trans_list, entry)
+		if (trans->msc_a == msc_a)
 			return trans;
 
 	return NULL;
@@ -208,19 +269,53 @@ struct gsm_trans *trans_has_conn(const struct gsm_subscriber_connection *conn)
  * facilities, which will then send the necessary release indications.
  * \param[in] conn Connection that is going to be closed.
  */
-void trans_conn_closed(struct gsm_subscriber_connection *conn)
+void trans_conn_closed(const struct msc_a *msc_a)
 {
-	struct gsm_trans *trans;
-
 	/* As part of the CC REL_IND the remote leg might be released and this
 	 * will trigger another call to trans_free. This is something the llist
 	 * macro can not handle and we need to re-iterate the list every time.
 	 */
-restart:
-	llist_for_each_entry(trans, &conn->network->trans_list, entry) {
-		if (trans->conn == conn) {
-			trans_free(trans);
-			goto restart;
-		}
+	struct gsm_trans *trans;
+	while ((trans = trans_has_conn(msc_a)))
+		trans_free(trans);
+}
+
+const struct value_string trans_type_names[] = {
+	{ TRANS_CC, "CC" },
+	{ TRANS_SMS, "SMS" },
+	{ TRANS_USSD, "NCSS" },
+	{ TRANS_SILENT_CALL, "silent-call" },
+	{}
+};
+
+uint8_t trans_type_to_gsm48_proto(enum trans_type type)
+{
+	switch (type) {
+	case TRANS_CC:
+	case TRANS_SILENT_CALL:
+		return GSM48_PDISC_CC;
+	case TRANS_SMS:
+		return GSM48_PDISC_SMS;
+	case TRANS_USSD:
+		return GSM48_PDISC_NC_SS;
+	default:
+		return GSM48_PDISC_TEST;
+	}
+
+}
+
+const char *trans_name(const struct gsm_trans *trans)
+{
+	static char namebuf[32];
+	if (!trans)
+		return "NULL";
+	switch (trans->type) {
+	case TRANS_CC:
+		snprintf(namebuf, sizeof(namebuf), "%s:%s",
+			 trans_type_name(trans->type), gsm48_cc_state_name(trans->cc.state));
+		return namebuf;
+
+	default:
+		return trans_type_name(trans->type);
 	}
 }
