@@ -60,6 +60,7 @@ struct proc_arq_priv {
 	void *parent_event_data;
 
 	enum vlr_parq_type type;
+	enum osmo_cm_service_type cm_service_type;
 	enum gsm48_reject_value result; /*< 0 on success */
 	bool by_tmsi;
 	char imsi[16];
@@ -67,12 +68,13 @@ struct proc_arq_priv {
 	struct osmo_location_area_id lai;
 	bool authentication_required;
 	bool ciphering_required;
+	uint8_t key_seq;
 	bool is_r99;
 	bool is_utran;
 	bool implicitly_accepted_parq_by_ciphering_cmd;
 };
 
-static void assoc_par_with_subscr(struct osmo_fsm_inst *fi, struct vlr_subscr *vsub)
+static int assoc_par_with_subscr(struct osmo_fsm_inst *fi, struct vlr_subscr *vsub)
 {
 	struct proc_arq_priv *par = fi->priv;
 	struct vlr_instance *vlr = par->vlr;
@@ -81,7 +83,7 @@ static void assoc_par_with_subscr(struct osmo_fsm_inst *fi, struct vlr_subscr *v
 	par->vsub = vsub;
 	/* Tell MSC to associate this subscriber with the given
 	 * connection */
-	vlr->ops.subscr_assoc(par->msc_conn_ref, par->vsub);
+	return vlr->ops.subscr_assoc(par->msc_conn_ref, par->vsub);
 }
 
 static const char *vlr_proc_arq_result_name(const struct osmo_fsm_inst *fi)
@@ -120,7 +122,8 @@ static void proc_arq_vlr_dispatch_result(struct osmo_fsm_inst *fi,
 	if (par->type == VLR_PR_ARQ_T_CM_SERV_REQ) {
 		if (success
 		    && !par->implicitly_accepted_parq_by_ciphering_cmd) {
-			rc = par->vlr->ops.tx_cm_serv_acc(par->msc_conn_ref);
+			rc = par->vlr->ops.tx_cm_serv_acc(par->msc_conn_ref,
+							  par->cm_service_type);
 			if (rc) {
 				LOGPFSML(fi, LOGL_ERROR,
 					 "Failed to send CM Service Accept\n");
@@ -129,6 +132,7 @@ static void proc_arq_vlr_dispatch_result(struct osmo_fsm_inst *fi,
 		}
 		if (!success) {
 			rc = par->vlr->ops.tx_cm_serv_rej(par->msc_conn_ref,
+							  par->cm_service_type,
 							  par->result);
 			if (rc)
 				LOGPFSML(fi, LOGL_ERROR,
@@ -140,7 +144,7 @@ static void proc_arq_vlr_dispatch_result(struct osmo_fsm_inst *fi,
 	 * will start handling pending paging transactions. */
 
 	if (!fi->proc.parent) {
-		LOGPFSML(fi, LOGL_ERROR, "No parent FSM");
+		LOGPFSML(fi, LOGL_ERROR, "No parent FSM\n");
 		return;
 	}
 	osmo_fsm_inst_dispatch(fi->proc.parent,
@@ -316,7 +320,8 @@ static bool is_auth_required(struct proc_arq_priv *par)
 	/* The cases where the authentication procedure should be used
 	 * are defined in 3GPP TS 33.102 */
 	/* For now we use a default value passed in to vlr_lu_fsm(). */
-	return par->authentication_required || par->ciphering_required;
+	return par->authentication_required ||
+		(par->ciphering_required && !auth_try_reuse_tuple(par->vsub, par->key_seq));
 }
 
 /* after the IMSI is known */
@@ -354,10 +359,10 @@ static void proc_arq_vlr_fn_init(struct osmo_fsm_inst *fi,
 	/* Obtain_Identity_VLR */
 	if (!par->by_tmsi) {
 		/* IMSI was included */
-		vsub = vlr_subscr_find_by_imsi(par->vlr, par->imsi);
+		vsub = vlr_subscr_find_by_imsi(par->vlr, par->imsi, __func__);
 	} else {
 		/* TMSI was included */
-		vsub = vlr_subscr_find_by_tmsi(par->vlr, par->tmsi);
+		vsub = vlr_subscr_find_by_tmsi(par->vlr, par->tmsi, __func__);
 	}
 	if (vsub) {
 		log_set_context(LOG_CTX_VLR_SUBSCR, vsub);
@@ -371,9 +376,11 @@ static void proc_arq_vlr_fn_init(struct osmo_fsm_inst *fi,
 					  GSM48_REJECT_NETWORK_FAILURE);
 		}
 		vsub->proc_arq_fsm = fi;
-		assoc_par_with_subscr(fi, vsub);
-		proc_arq_vlr_fn_post_imsi(fi);
-		vlr_subscr_put(vsub);
+		if (assoc_par_with_subscr(fi, vsub) != 0)
+			proc_arq_fsm_done(fi, GSM48_REJECT_NETWORK_FAILURE);
+		else
+			proc_arq_vlr_fn_post_imsi(fi);
+		vlr_subscr_put(vsub, __func__);
 		return;
 	}
 	/* No VSUB could be resolved. What now? */
@@ -381,7 +388,7 @@ static void proc_arq_vlr_fn_init(struct osmo_fsm_inst *fi,
 	if (!par->by_tmsi) {
 		/* We couldn't find a subscriber even by IMSI,
 		 * Set User Error: Unidentified Subscriber */
-		proc_arq_fsm_done(fi, GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE);
+		proc_arq_fsm_done(fi, GSM48_REJECT_IMSI_UNKNOWN_IN_VLR);
 		return;
 	} else {
 		/* TMSI was included, are we permitted to use it? */
@@ -392,7 +399,7 @@ static void proc_arq_vlr_fn_init(struct osmo_fsm_inst *fi,
 			return;
 		} else {
 			/* Set User Error: Unidentified Subscriber */
-			proc_arq_fsm_done(fi, GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE);
+			proc_arq_fsm_done(fi, GSM48_REJECT_IMSI_UNKNOWN_IN_VLR);
 			return;
 		}
 	}
@@ -408,15 +415,17 @@ static void proc_arq_vlr_fn_w_obt_imsi(struct osmo_fsm_inst *fi,
 
 	OSMO_ASSERT(event == PR_ARQ_E_ID_IMSI);
 
-	vsub = vlr_subscr_find_by_imsi(vlr, par->imsi);
+	vsub = vlr_subscr_find_by_imsi(vlr, par->imsi, __func__);
 	if (!vsub) {
 		/* Set User Error: Unidentified Subscriber */
-		proc_arq_fsm_done(fi, GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE);
+		proc_arq_fsm_done(fi, GSM48_REJECT_IMSI_UNKNOWN_IN_VLR);
 		return;
 	}
-	assoc_par_with_subscr(fi, vsub);
-	proc_arq_vlr_fn_post_imsi(fi);
-	vlr_subscr_put(vsub);
+	if (assoc_par_with_subscr(fi, vsub))
+		proc_arq_fsm_done(fi, GSM48_REJECT_NETWORK_FAILURE);
+	else
+		proc_arq_vlr_fn_post_imsi(fi);
+	vlr_subscr_put(vsub, __func__);
 }
 
 /* Authenticate_VLR has completed */
@@ -439,37 +448,28 @@ static void proc_arq_vlr_fn_w_auth(struct osmo_fsm_inst *fi,
 static void proc_arq_vlr_fn_w_ciph(struct osmo_fsm_inst *fi,
 				   uint32_t event, void *data)
 {
-	struct proc_arq_priv *par = fi->priv;
-	struct vlr_subscr *vsub = par->vsub;
-	struct vlr_ciph_result res = { .cause = VLR_CIPH_REJECT };
+	enum vlr_ciph_result_cause result = VLR_CIPH_REJECT;
 
 	OSMO_ASSERT(event == PR_ARQ_E_CIPH_RES);
 
 	if (!data)
 		LOGPFSML(fi, LOGL_ERROR, "invalid ciphering result: NULL\n");
 	else
-		res = *(struct vlr_ciph_result*)data;
+		result = *(enum vlr_ciph_result_cause*)data;
 
-	switch (res.cause) {
+	switch (result) {
 	case VLR_CIPH_COMPL:
-		break;
+		_proc_arq_vlr_node2_post_ciph(fi);
+		return;
 	case VLR_CIPH_REJECT:
 		LOGPFSM(fi, "ciphering rejected\n");
 		proc_arq_fsm_done(fi, GSM48_REJECT_ILLEGAL_MS);
 		return;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "invalid ciphering result: %d\n",
-			 res.cause);
+		LOGPFSML(fi, LOGL_ERROR, "invalid ciphering result: %d\n", result);
 		proc_arq_fsm_done(fi, GSM48_REJECT_ILLEGAL_MS);
 		return;
 	}
-
-
-	if (*res.imeisv) {
-		LOGPFSM(fi, "got IMEISV: %s\n", res.imeisv);
-		vlr_subscr_set_imeisv(vsub, res.imeisv);
-	}
-	_proc_arq_vlr_node2_post_ciph(fi);
 }
 
 /* Update_Location_Child_VLR has completed */
@@ -524,6 +524,7 @@ static const struct osmo_fsm_state proc_arq_vlr_states[] = {
 		.out_state_mask = S(PR_ARQ_S_DONE) |
 				  S(PR_ARQ_S_WAIT_OBTAIN_IMSI) |
 				  S(PR_ARQ_S_WAIT_AUTH) |
+				  S(PR_ARQ_S_WAIT_CIPH) |
 				  S(PR_ARQ_S_WAIT_UPD_LOC_CHILD) |
 				  S(PR_ARQ_S_WAIT_SUB_PRES) |
 				  S(PR_ARQ_S_WAIT_TRACE_SUB) |
@@ -536,6 +537,7 @@ static const struct osmo_fsm_state proc_arq_vlr_states[] = {
 		.in_event_mask = S(PR_ARQ_E_ID_IMSI),
 		.out_state_mask = S(PR_ARQ_S_DONE) |
 				  S(PR_ARQ_S_WAIT_AUTH) |
+				  S(PR_ARQ_S_WAIT_CIPH) |
 				  S(PR_ARQ_S_WAIT_UPD_LOC_CHILD) |
 				  S(PR_ARQ_S_WAIT_SUB_PRES) |
 				  S(PR_ARQ_S_WAIT_TRACE_SUB) |
@@ -629,10 +631,12 @@ vlr_proc_acc_req(struct osmo_fsm_inst *parent,
 		 uint32_t parent_event_failure,
 		 void *parent_event_data,
 		 struct vlr_instance *vlr, void *msc_conn_ref,
-		 enum vlr_parq_type type, const uint8_t *mi_lv,
+		 enum vlr_parq_type type, enum osmo_cm_service_type cm_service_type,
+		 const uint8_t *mi_lv,
 		 const struct osmo_location_area_id *lai,
 		 bool authentication_required,
 		 bool ciphering_required,
+		 uint8_t key_seq,
 		 bool is_r99, bool is_utran)
 {
 	struct osmo_fsm_inst *fi;
@@ -650,12 +654,14 @@ vlr_proc_acc_req(struct osmo_fsm_inst *parent,
 	par->vlr = vlr;
 	par->msc_conn_ref = msc_conn_ref;
 	par->type = type;
+	par->cm_service_type = cm_service_type;
 	par->lai = *lai;
 	par->parent_event_success = parent_event_success;
 	par->parent_event_failure = parent_event_failure;
 	par->parent_event_data = parent_event_data;
 	par->authentication_required = authentication_required;
 	par->ciphering_required = ciphering_required;
+	par->key_seq = key_seq;
 	par->is_r99 = is_r99;
 	par->is_utran = is_utran;
 
@@ -676,8 +682,7 @@ vlr_proc_acc_req(struct osmo_fsm_inst *parent,
 	mi_type = mi_lv[1] & GSM_MI_TYPE_MASK;
 	switch (mi_type) {
 	case GSM_MI_TYPE_IMSI:
-		strncpy(par->imsi, mi_string, sizeof(par->imsi)-1);
-		par->imsi[sizeof(par->imsi)-1] = '\0';
+		osmo_strlcpy(par->imsi, mi_string, sizeof(par->imsi));
 		par->by_tmsi = false;
 		break;
 	case GSM_MI_TYPE_TMSI:
@@ -687,7 +692,7 @@ vlr_proc_acc_req(struct osmo_fsm_inst *parent,
 	case GSM_MI_TYPE_IMEI:
 		/* TODO: IMEI (emergency call) */
 	default:
-		proc_arq_fsm_done(fi, GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE);
+		proc_arq_fsm_done(fi, GSM48_REJECT_INVALID_MANDANTORY_INF);
 		return;
 	}
 
@@ -770,6 +775,6 @@ static struct osmo_fsm upd_loc_child_vlr_fsm = {
 
 void vlr_parq_fsm_init(void)
 {
-	//osmo_fsm_register(&upd_loc_child_vlr_fsm);
-	osmo_fsm_register(&proc_arq_vlr_fsm);
+	//OSMO_ASSERT(osmo_fsm_register(&upd_loc_child_vlr_fsm) == 0);
+	OSMO_ASSERT(osmo_fsm_register(&proc_arq_vlr_fsm) == 0);
 }
