@@ -27,6 +27,7 @@
 #include "vlr_core.h"
 #include "vlr_auth_fsm.h"
 #include "vlr_lu_fsm.h"
+#include "vlr_sgs_fsm.h"
 
 #define S(x)	(1 << (x))
 
@@ -79,7 +80,7 @@ static void upd_hlr_vlr_fsm_init(struct osmo_fsm_inst *fi, uint32_t event,
 	OSMO_ASSERT(event == UPD_HLR_VLR_E_START);
 
 	/* Send UpdateLocation to HLR */
-	rc = vlr_subscr_req_lu(vsub, vsub->vlr->cfg.is_ps);
+	rc = vlr_subscr_req_lu(vsub);
 	if (rc < 0)
 		LOGPFSML(fi, LOGL_ERROR, "Failed to send UpdateLocation to HLR\n");
 	osmo_fsm_inst_state_chg(fi, UPD_HLR_VLR_S_WAIT_FOR_DATA,
@@ -271,7 +272,9 @@ static inline struct vlr_subscr *sub_pres_vlr_fi_priv(struct osmo_fsm_inst *fi)
 	return (struct vlr_subscr*)fi->priv;
 }
 
-/* Note that the start event is dispatched right away, so in case the FSM immediately concludes from that
+/* THIS IS CURRENTLY DEAD CODE, SINCE WE NEVER SET vsub->ms_not_reachable_flag = true.
+ *
+ * Note that the start event is dispatched right away, so in case the FSM immediately concludes from that
  * event, the created FSM struct may no longer be valid as it already deallocated again, and it may
  * furthermore already have invoked the parent FSM instance's deallocation as well. Hence, instead of
  * returning, store the created FSM instance address in *fi_p before dispatching the event. It is thus
@@ -356,17 +359,11 @@ static void vlr_lu_compl_fsm_success(struct osmo_fsm_inst *fi)
 	struct vlr_subscr *vsub = lcvp->vsub;
 	if (!vsub->lu_complete) {
 		vsub->lu_complete = true;
-		/* Balanced by vlr_subscr_rx_imsi_detach() */
-		vlr_subscr_get(vsub);
+		/* Balanced by vlr_subscr_expire() */
+		vlr_subscr_get(vsub, VSUB_USE_ATTACHED);
 	}
 	_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_SUCCESS, 0);
-}
-
-static void vlr_lu_compl_fsm_failure(struct osmo_fsm_inst *fi, uint8_t cause)
-{
-	struct lu_compl_vlr_priv *lcvp = lu_compl_vlr_fi_priv(fi);
-	lcvp->vsub->vlr->ops.tx_lu_rej(lcvp->msc_conn_ref, cause);
-	_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_FAILURE, cause);
+	vlr_sgs_fsm_update_id(vsub);
 }
 
 static void vlr_lu_compl_fsm_dispatch_result(struct osmo_fsm_inst *fi,
@@ -413,7 +410,12 @@ static void lu_compl_vlr_init(struct osmo_fsm_inst *fi, uint32_t event,
 	osmo_fsm_inst_state_chg(fi, LU_COMPL_VLR_S_WAIT_SUB_PRES,
 				LU_TIMEOUT_LONG, 0);
 
-	sub_pres_vlr_fsm_start(&lcvp->sub_pres_vlr_fsm, fi, vsub, LU_COMPL_VLR_E_SUB_PRES_COMPL);
+	/* If ms_not_reachable_flag == false, the sub_pres_vlr_fsm will anyway terminate straight away and dispatch
+	 * LU_COMPL_VLR_E_SUB_PRES_COMPL to this fi, so we might as well skip that dance here and save some logging. */
+	if (vsub->ms_not_reachable_flag)
+		sub_pres_vlr_fsm_start(&lcvp->sub_pres_vlr_fsm, fi, vsub, LU_COMPL_VLR_E_SUB_PRES_COMPL);
+	else
+		osmo_fsm_inst_dispatch(fi, LU_COMPL_VLR_E_SUB_PRES_COMPL, NULL);
 }
 
 static void lu_compl_vlr_new_tmsi(struct osmo_fsm_inst *fi)
@@ -425,8 +427,7 @@ static void lu_compl_vlr_new_tmsi(struct osmo_fsm_inst *fi)
 	LOGPFSM(fi, "%s()\n", __func__);
 
 	if (vlr_subscr_alloc_tmsi(vsub)) {
-		vlr_lu_compl_fsm_failure(fi,
-					 GSM48_REJECT_SRV_OPT_TMP_OUT_OF_ORDER);
+		_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_FAILURE, GSM48_REJECT_SRV_OPT_TMP_OUT_OF_ORDER);
 		return;
 	}
 
@@ -451,7 +452,8 @@ static void lu_compl_vlr_wait_subscr_pres(struct osmo_fsm_inst *fi,
 
 	/* TODO: Trace_Subscriber_Activity_VLR */
 
-	if (vlr->cfg.check_imei_rqd) {
+	/* If imeisv_early is enabled: IMEI already retrieved and checked (vlr_loc_upd_node1_pre), don't do it again. */
+	if (vlr->cfg.check_imei_rqd && !vlr->cfg.retrieve_imeisv_early) {
 		/* Check IMEI VLR */
 		osmo_fsm_inst_state_chg(fi,
 					lcvp->assign_tmsi ?
@@ -485,15 +487,14 @@ static void lu_compl_vlr_wait_imei(struct osmo_fsm_inst *fi, uint32_t event,
 	case LU_COMPL_VLR_E_IMEI_CHECK_ACK:
 		if (!vsub->imei[0]) {
 			/* Abort: Do nothing */
-			vlr_lu_compl_fsm_failure(fi,
-						 GSM48_REJECT_PROTOCOL_ERROR);
+			_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_FAILURE, GSM48_REJECT_PROTOCOL_ERROR);
 			return;
 		}
 		/* Pass */
 		break;
 
 	case LU_COMPL_VLR_E_IMEI_CHECK_NACK:
-		vlr_lu_compl_fsm_failure(fi, GSM48_REJECT_ILLEGAL_ME);
+		_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_FAILURE, GSM48_REJECT_ILLEGAL_ME);
 		/* FIXME: IMEI Check Fail to VLR Application (Detach IMSI VLR) */
 		return;
 	}
@@ -534,12 +535,13 @@ static void lu_compl_vlr_wait_tmsi(struct osmo_fsm_inst *fi, uint32_t event,
 		LOGPFSML(fi, LOGL_ERROR, "TMSI Realloc Compl implies that"
 			 " the subscriber has a new TMSI allocated, but"
 			 " the new TMSI is unset.\n");
-		vlr_lu_compl_fsm_failure(fi, GSM48_REJECT_NETWORK_FAILURE);
+		_vlr_lu_compl_fsm_done(fi, VLR_FSM_RESULT_FAILURE, GSM48_REJECT_NETWORK_FAILURE);
 		return;
 	}
 
 	vsub->tmsi = vsub->tmsi_new;
 	vsub->tmsi_new = GSM_RESERVED_TMSI;
+	vsub->vlr->ops.subscr_update(vsub);
 
 	vlr_lu_compl_fsm_success(fi);
 }
@@ -548,8 +550,7 @@ static const struct osmo_fsm_state lu_compl_vlr_states[] = {
 	[LU_COMPL_VLR_S_INIT] = {
 		.in_event_mask = S(LU_COMPL_VLR_E_START),
 		.out_state_mask = S(LU_COMPL_VLR_S_DONE) |
-				  S(LU_COMPL_VLR_S_WAIT_SUB_PRES) |
-				  S(LU_COMPL_VLR_S_WAIT_IMEI),
+				  S(LU_COMPL_VLR_S_WAIT_SUB_PRES),
 		.name = OSMO_STRINGIFY(LU_COMPL_VLR_S_INIT),
 		.action = lu_compl_vlr_init,
 	},
@@ -646,6 +647,8 @@ static const struct value_string fsm_lu_event_names[] = {
 	OSMO_VALUE_STRING(VLR_ULA_E_ID_IMSI),
 	OSMO_VALUE_STRING(VLR_ULA_E_ID_IMEI),
 	OSMO_VALUE_STRING(VLR_ULA_E_ID_IMEISV),
+	OSMO_VALUE_STRING(VLR_ULA_E_HLR_IMEI_ACK),
+	OSMO_VALUE_STRING(VLR_ULA_E_HLR_IMEI_NACK),
 	OSMO_VALUE_STRING(VLR_ULA_E_HLR_LU_RES),
 	OSMO_VALUE_STRING(VLR_ULA_E_UPD_HLR_COMPL),
 	OSMO_VALUE_STRING(VLR_ULA_E_LU_COMPL_SUCCESS),
@@ -674,6 +677,7 @@ struct lu_fsm_priv {
 	struct osmo_location_area_id new_lai;
 	bool authentication_required;
 	bool ciphering_required;
+	uint8_t key_seq;
 	bool is_r99;
 	bool is_utran;
 	bool assign_tmsi;
@@ -684,7 +688,7 @@ struct lu_fsm_priv {
 static bool lai_in_this_vlr(struct vlr_instance *vlr,
 			    const struct osmo_location_area_id *lai)
 {
-	/* TODO: VLR needs to keep a locally configued list of LAIs */
+	/* TODO: VLR needs to keep a locally configured list of LAIs */
 	return true;
 }
 
@@ -694,7 +698,8 @@ static bool is_auth_required(struct lu_fsm_priv *lfp)
 	/* The cases where the authentication procedure should be used
 	 * are defined in 3GPP TS 33.102 */
 	/* For now we use a default value passed in to vlr_lu_fsm(). */
-	return lfp->authentication_required || lfp->ciphering_required;
+	return lfp->authentication_required ||
+		(lfp->ciphering_required && !auth_try_reuse_tuple(lfp->vsub, lfp->key_seq));
 }
 
 /* Determine if ciphering is required */
@@ -911,6 +916,39 @@ static void vlr_loc_upd_node1(struct osmo_fsm_inst *fi)
 	}
 }
 
+static void vlr_loc_upd_node1_pre(struct osmo_fsm_inst *fi)
+{
+	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
+	struct vlr_instance *vlr = lfp->vlr;
+
+	LOGPFSM(fi, "%s()\n", __func__);
+
+	if (vlr->cfg.check_imei_rqd && vlr->cfg.retrieve_imeisv_early) {
+		osmo_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY, vlr_timer(lfp->vlr, 3270), 3270);
+		vlr_subscr_tx_req_check_imei(lfp->vsub);
+	} else {
+		vlr_loc_upd_node1(fi);
+	}
+}
+
+/* End of Check_IMEI Procedure. Executed early (before the location update), so we can send the IMEI to the HLR even if
+ * the MS would be rejected in LU. See the "Configuring the Subscribers Create on Demand Feature" section of the OsmoHLR
+ * user manual for a detailed explanation of the use case. */
+static void lu_fsm_wait_hlr_check_imei_early(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+{
+	switch (event) {
+	case VLR_ULA_E_HLR_IMEI_ACK:
+		vlr_loc_upd_node1(fi);
+		break;
+	case VLR_ULA_E_HLR_IMEI_NACK:
+		lu_fsm_failure(fi, GSM48_REJECT_ILLEGAL_ME);
+		break;
+	default:
+		OSMO_ASSERT(0);
+		break;
+	}
+}
+
 static void vlr_loc_upd_want_imsi(struct osmo_fsm_inst *fi)
 {
 	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
@@ -924,7 +962,7 @@ static void vlr_loc_upd_want_imsi(struct osmo_fsm_inst *fi)
 	osmo_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_IMSI,
 				vlr_timer(vlr, 3270), 3270);
 	vlr->ops.tx_id_req(lfp->msc_conn_ref, GSM_MI_TYPE_IMSI);
-	/* will continue at vlr_loc_upd_node1() once IMSI arrives */
+	/* will continue at vlr_loc_upd_node1_pre() once IMSI arrives */
 }
 
 static int assoc_lfp_with_sub(struct osmo_fsm_inst *fi, struct vlr_subscr *vsub)
@@ -944,11 +982,12 @@ static int assoc_lfp_with_sub(struct osmo_fsm_inst *fi, struct vlr_subscr *vsub)
 	vsub->lu_fsm = fi;
 	vsub->msc_conn_ref = lfp->msc_conn_ref;
 	/* FIXME: send new LAC to HLR? */
-	vsub->lac = lfp->new_lai.lac;
+	vsub->cgi.lai.lac = lfp->new_lai.lac;
 	lfp->vsub = vsub;
 	/* Tell MSC to associate this subscriber with the given
 	 * connection */
-	vlr->ops.subscr_assoc(lfp->msc_conn_ref, lfp->vsub);
+	if (vlr->ops.subscr_assoc(lfp->msc_conn_ref, lfp->vsub))
+		lu_fsm_failure(fi, GSM48_REJECT_NETWORK_FAILURE);
 	return 0;
 }
 
@@ -968,7 +1007,7 @@ static int _lu_fsm_associate_vsub(struct osmo_fsm_inst *fi)
 		 * someone else's TMSI.
 		 * TODO: Otherwise we can ask for the IMSI and verify that it
 		 * matches the IMSI on record. */
-		vsub = vlr_subscr_find_or_create_by_tmsi(vlr, lfp->tmsi, NULL);
+		vsub = vlr_subscr_find_or_create_by_tmsi(vlr, lfp->tmsi, __func__, NULL);
 
 		if (!vsub) {
 			LOGPFSML(fi, LOGL_ERROR, "VLR subscriber allocation failed\n");
@@ -978,27 +1017,27 @@ static int _lu_fsm_associate_vsub(struct osmo_fsm_inst *fi)
 
 		vsub->sub_dataconf_by_hlr_ind = false;
 		if (assoc_lfp_with_sub(fi, vsub)) {
-			vlr_subscr_put(vsub);
+			vlr_subscr_put(vsub, __func__);
 			return -1; /* error, fsm failure invoked in assoc_lfp_with_sub() */
 		}
-		vlr_subscr_put(vsub);
+		vlr_subscr_put(vsub, __func__);
 	} else {
 		/* IMSI was used */
-		vsub = vlr_subscr_find_or_create_by_imsi(vlr, lfp->imsi, NULL);
+		vsub = vlr_subscr_find_or_create_by_imsi(vlr, lfp->imsi, __func__, NULL);
 
 		if (!vsub) {
 			LOGPFSML(fi, LOGL_ERROR, "VLR subscriber allocation failed\n");
 			lu_fsm_failure(fi, GSM48_REJECT_SRV_OPT_TMP_OUT_OF_ORDER);
-			vlr_subscr_put(vsub);
+			vlr_subscr_put(vsub, __func__);
 			return -1;
 		}
 
 		vsub->sub_dataconf_by_hlr_ind = false;
 		if (assoc_lfp_with_sub(fi, vsub)) {
-			vlr_subscr_put(vsub);
+			vlr_subscr_put(vsub, __func__);
 			return -1; /* error, fsm failure invoked in assoc_lfp_with_sub() */
 		}
-		vlr_subscr_put(vsub);
+		vlr_subscr_put(vsub, __func__);
 	}
 	return 0;
 }
@@ -1029,7 +1068,7 @@ static void _start_lu_main(struct osmo_fsm_inst *fi)
 	if (!lfp->vsub->imsi[0])
 		vlr_loc_upd_want_imsi(fi);
 	else
-		vlr_loc_upd_node1(fi);
+		vlr_loc_upd_node1_pre(fi);
 }
 
 static void lu_fsm_idle(struct osmo_fsm_inst *fi, uint32_t event,
@@ -1045,6 +1084,11 @@ static void lu_fsm_idle(struct osmo_fsm_inst *fi, uint32_t event,
 
 	OSMO_ASSERT(lfp->vsub);
 
+	/* At this point we know for which subscriber the location update is,
+	 * we now must inform SGs-UE FSM that we received a location update
+	 * via A, IU or Gs interface. */
+	osmo_fsm_inst_dispatch(lfp->vsub->sgs_fsm, SGS_UE_E_RX_LU_FROM_A_IU_GS, NULL);
+
 	/* See 3GPP TS 23.012, procedure Retrieve_IMEISV_If_Required */
 	if ((!vlr->cfg.retrieve_imeisv_early)
 	    || (lfp->type == VLR_LU_TYPE_PERIODIC && lfp->vsub->imeisv[0])) {
@@ -1067,8 +1111,7 @@ static void lu_fsm_wait_imeisv(struct osmo_fsm_inst *fi, uint32_t event,
 		_start_lu_main(fi);
 		break;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-			 osmo_fsm_event_name(fi->fsm, event));
+		OSMO_ASSERT(0);
 		break;
 	}
 }
@@ -1079,14 +1122,13 @@ static void lu_fsm_wait_pvlr(struct osmo_fsm_inst *fi, uint32_t event,
 {
 	switch (event) {
 	case VLR_ULA_E_SEND_ID_ACK:
-		vlr_loc_upd_node1(fi);
+		vlr_loc_upd_node1_pre(fi);
 		break;
 	case VLR_ULA_E_SEND_ID_NACK:
 		vlr_loc_upd_want_imsi(fi);
 		break;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-			 osmo_fsm_event_name(fi->fsm, event));
+		OSMO_ASSERT(0);
 		break;
 	}
 }
@@ -1114,36 +1156,28 @@ static void lu_fsm_wait_auth(struct osmo_fsm_inst *fi, uint32_t event,
 static void lu_fsm_wait_ciph(struct osmo_fsm_inst *fi, uint32_t event,
 			     void *data)
 {
-	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
-	struct vlr_subscr *vsub = lfp->vsub;
-	struct vlr_ciph_result res = { .cause = VLR_CIPH_REJECT };
+	enum vlr_ciph_result_cause result = VLR_CIPH_REJECT;
 
 	OSMO_ASSERT(event == VLR_ULA_E_CIPH_RES);
 
 	if (!data)
 		LOGPFSML(fi, LOGL_ERROR, "invalid ciphering result: NULL\n");
 	else
-		res = *(struct vlr_ciph_result*)data;
+		result = *(enum vlr_ciph_result_cause*)data;
 
-	switch (res.cause) {
+	switch (result) {
 	case VLR_CIPH_COMPL:
-		break;
+		vlr_loc_upd_post_ciph(fi);
+		return;
 	case VLR_CIPH_REJECT:
 		LOGPFSM(fi, "ciphering rejected\n");
 		lu_fsm_failure(fi, GSM48_REJECT_INVALID_MANDANTORY_INF);
 		return;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "invalid ciphering result: %d\n",
-			 res.cause);
+		LOGPFSML(fi, LOGL_ERROR, "invalid ciphering result: %d\n", result);
 		lu_fsm_failure(fi, GSM48_REJECT_INVALID_MANDANTORY_INF);
 		return;
 	}
-
-	if (*res.imeisv) {
-		LOGPFSM(fi, "got IMEISV: %s\n", res.imeisv);
-		vlr_subscr_set_imeisv(vsub, res.imeisv);
-	}
-	vlr_loc_upd_post_ciph(fi);
 }
 
 static void lu_fsm_wait_imsi(struct osmo_fsm_inst *fi, uint32_t event,
@@ -1156,11 +1190,10 @@ static void lu_fsm_wait_imsi(struct osmo_fsm_inst *fi, uint32_t event,
 	switch (event) {
 	case VLR_ULA_E_ID_IMSI:
 		vlr_subscr_set_imsi(vsub, mi_string);
-		vlr_loc_upd_node1(fi);
+		vlr_loc_upd_node1_pre(fi);
 		break;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-			 osmo_fsm_event_name(fi->fsm, event));
+		OSMO_ASSERT(0);
 		break;
 	}
 }
@@ -1188,8 +1221,8 @@ static void lu_fsm_wait_hlr_ul_res(struct osmo_fsm_inst *fi, uint32_t event,
 			/* continue in MSC ?!? */
 		} else {
 			/* unsuccessful case */
-			enum gsm48_gmm_cause cause =
-				*(enum gsm48_gmm_cause *)data;
+			enum gsm48_reject_value cause =
+				*(enum gsm48_reject_value *)data;
 			/* Ignoring standalone mode for now. */
 			if (0 /* procedure_error && vlr->cfg.standalone_mode */) {
 				osmo_fsm_inst_state_chg(fi,
@@ -1201,9 +1234,12 @@ static void lu_fsm_wait_hlr_ul_res(struct osmo_fsm_inst *fi, uint32_t event,
 			}
 		}
 		break;
+	case VLR_ULA_E_ID_IMEI:
+	case VLR_ULA_E_ID_IMEISV:
+		/* Got the IMEI from ME, nothing to do right now though. */
+		break;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-			 osmo_fsm_event_name(fi->fsm, event));
+		OSMO_ASSERT(0);
 		break;
 	}
 }
@@ -1221,9 +1257,18 @@ static void lu_fsm_wait_lu_compl(struct osmo_fsm_inst *fi, uint32_t event,
 				       LU_COMPL_VLR_E_NEW_TMSI_ACK, NULL);
 		break;
 	case VLR_ULA_E_ID_IMEI:
+	case VLR_ULA_E_ID_IMEISV:
+		/* Got the IMEI from ME, now send it to HLR */
+		vlr_subscr_tx_req_check_imei(lfp->vsub);
+		break;
+	case VLR_ULA_E_HLR_IMEI_ACK:
 		osmo_fsm_inst_dispatch(lfp->lu_compl_vlr_fsm,
 				       LU_COMPL_VLR_E_IMEI_CHECK_ACK, NULL);
 		break;
+	case VLR_ULA_E_HLR_IMEI_NACK:
+		osmo_fsm_inst_dispatch(lfp->lu_compl_vlr_fsm,
+				       LU_COMPL_VLR_E_IMEI_CHECK_NACK, NULL);
+		break;
 	case VLR_ULA_E_LU_COMPL_SUCCESS:
 		lu_fsm_discard_lu_compl_fsm(fi);
 
@@ -1245,8 +1290,7 @@ static void lu_fsm_wait_lu_compl(struct osmo_fsm_inst *fi, uint32_t event,
 		lu_fsm_failure(fi, cause);
 		break;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-			 osmo_fsm_event_name(fi->fsm, event));
+		OSMO_ASSERT(0);
 		break;
 	}
 }
@@ -1278,8 +1322,7 @@ static void lu_fsm_wait_lu_compl_standalone(struct osmo_fsm_inst *fi,
 		lu_fsm_failure(fi, cause);
 		break;
 	default:
-		LOGPFSML(fi, LOGL_ERROR, "event without effect: %s\n",
-			 osmo_fsm_event_name(fi->fsm, event));
+		OSMO_ASSERT(0);
 		break;
 	}
 }
@@ -1291,6 +1334,8 @@ static const struct osmo_fsm_state vlr_lu_fsm_states[] = {
 				  S(VLR_ULA_S_WAIT_PVLR) |
 				  S(VLR_ULA_S_WAIT_IMSI) |
 				  S(VLR_ULA_S_WAIT_AUTH) |
+				  S(VLR_ULA_S_WAIT_CIPH) |
+				  S(VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY) |
 				  S(VLR_ULA_S_WAIT_HLR_UPD) |
 				  S(VLR_ULA_S_DONE),
 		.name = OSMO_STRINGIFY(VLR_ULA_S_IDLE),
@@ -1301,6 +1346,8 @@ static const struct osmo_fsm_state vlr_lu_fsm_states[] = {
 		.out_state_mask = S(VLR_ULA_S_WAIT_PVLR) |
 				  S(VLR_ULA_S_WAIT_IMSI) |
 				  S(VLR_ULA_S_WAIT_AUTH) |
+				  S(VLR_ULA_S_WAIT_CIPH) |
+				  S(VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY) |
 				  S(VLR_ULA_S_WAIT_HLR_UPD) |
 				  S(VLR_ULA_S_DONE),
 		.name = OSMO_STRINGIFY(VLR_ULA_S_WAIT_IMEISV),
@@ -1311,6 +1358,8 @@ static const struct osmo_fsm_state vlr_lu_fsm_states[] = {
 				 S(VLR_ULA_E_SEND_ID_NACK),
 		.out_state_mask = S(VLR_ULA_S_WAIT_IMSI) |
 				  S(VLR_ULA_S_WAIT_AUTH) |
+				  S(VLR_ULA_S_WAIT_CIPH) |
+				  S(VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY) |
 				  S(VLR_ULA_S_DONE),
 		.name = OSMO_STRINGIFY(VLR_ULA_S_WAIT_PVLR),
 		.action = lu_fsm_wait_pvlr,
@@ -1335,14 +1384,29 @@ static const struct osmo_fsm_state vlr_lu_fsm_states[] = {
 	[VLR_ULA_S_WAIT_IMSI] = {
 		.in_event_mask = S(VLR_ULA_E_ID_IMSI),
 		.out_state_mask = S(VLR_ULA_S_WAIT_AUTH) |
+				  S(VLR_ULA_S_WAIT_CIPH) |
+				  S(VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY) |
 				  S(VLR_ULA_S_WAIT_HLR_UPD) |
 				  S(VLR_ULA_S_DONE),
 		.name = OSMO_STRINGIFY(VLR_ULA_S_WAIT_IMSI),
 		.action = lu_fsm_wait_imsi,
 	},
+	[VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY] = {
+		.in_event_mask = S(VLR_ULA_E_HLR_IMEI_ACK) |
+				 S(VLR_ULA_E_HLR_IMEI_NACK),
+		.out_state_mask = S(VLR_ULA_S_WAIT_AUTH) |
+				  S(VLR_ULA_S_WAIT_CIPH) |
+				  S(VLR_ULA_S_WAIT_HLR_UPD) |
+				  S(VLR_ULA_S_WAIT_LU_COMPL) |
+				  S(VLR_ULA_S_DONE),
+		.name = OSMO_STRINGIFY(VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY),
+		.action = lu_fsm_wait_hlr_check_imei_early,
+	},
 	[VLR_ULA_S_WAIT_HLR_UPD] = {
 		.in_event_mask = S(VLR_ULA_E_HLR_LU_RES) |
-				 S(VLR_ULA_E_UPD_HLR_COMPL),
+				 S(VLR_ULA_E_UPD_HLR_COMPL) |
+				 S(VLR_ULA_E_ID_IMEI) |
+				 S(VLR_ULA_E_ID_IMEISV),
 		.out_state_mask = S(VLR_ULA_S_WAIT_LU_COMPL) |
 				  S(VLR_ULA_S_WAIT_LU_COMPL_STANDALONE) |
 				  S(VLR_ULA_S_DONE),
@@ -1354,7 +1418,9 @@ static const struct osmo_fsm_state vlr_lu_fsm_states[] = {
 				 S(VLR_ULA_E_LU_COMPL_FAILURE) |
 				 S(VLR_ULA_E_NEW_TMSI_ACK) |
 				 S(VLR_ULA_E_ID_IMEI) |
-				 S(VLR_ULA_E_ID_IMEISV),
+				 S(VLR_ULA_E_ID_IMEISV) |
+				 S(VLR_ULA_E_HLR_IMEI_ACK) |
+				 S(VLR_ULA_E_HLR_IMEI_NACK),
 		.out_state_mask = S(VLR_ULA_S_DONE),
 		.name = OSMO_STRINGIFY(VLR_ULA_S_WAIT_LU_COMPL),
 		.action = lu_fsm_wait_lu_compl,
@@ -1412,6 +1478,7 @@ vlr_loc_update(struct osmo_fsm_inst *parent,
 	       const struct osmo_location_area_id *new_lai,
 	       bool authentication_required,
 	       bool ciphering_required,
+	       uint8_t key_seq,
 	       bool is_r99, bool is_utran,
 	       bool assign_tmsi)
 {
@@ -1435,6 +1502,7 @@ vlr_loc_update(struct osmo_fsm_inst *parent,
 	lfp->parent_event_data = parent_event_data;
 	lfp->authentication_required = authentication_required;
 	lfp->ciphering_required = ciphering_required;
+	lfp->key_seq = key_seq;
 	lfp->is_r99 = is_r99;
 	lfp->is_utran = is_utran;
 	lfp->assign_tmsi = assign_tmsi;
@@ -1481,8 +1549,8 @@ void vlr_loc_update_cancel(struct osmo_fsm_inst *fi,
 
 void vlr_lu_fsm_init(void)
 {
-	osmo_fsm_register(&vlr_lu_fsm);
-	osmo_fsm_register(&upd_hlr_vlr_fsm);
-	osmo_fsm_register(&sub_pres_vlr_fsm);
-	osmo_fsm_register(&lu_compl_vlr_fsm);
+	OSMO_ASSERT(osmo_fsm_register(&vlr_lu_fsm) == 0);
+	OSMO_ASSERT(osmo_fsm_register(&upd_hlr_vlr_fsm) == 0);
+	OSMO_ASSERT(osmo_fsm_register(&sub_pres_vlr_fsm) == 0);
+	OSMO_ASSERT(osmo_fsm_register(&lu_compl_vlr_fsm) == 0);
 }
