@@ -1,7 +1,7 @@
 /* OsmoMSC - Circuit-Switched Core Network (MSC+VLR+HLR+SMSC) implementation
  */
 
-/* (C) 2016 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
+/* (C) 2016-2019 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * Based on OsmoNITB:
@@ -24,6 +24,7 @@
  *
  */
 
+#include <stdbool.h>
 #include <unistd.h>
 #include <time.h>
 #include <errno.h>
@@ -35,7 +36,7 @@
 #include <getopt.h>
 
 /* build switches from the configure script */
-#include "../../bscconfig.h"
+#include "config.h"
 
 #include <osmocom/msc/db.h>
 #include <osmocom/core/application.h>
@@ -46,7 +47,6 @@
 #include <osmocom/abis/e1_input.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/msc/signal.h>
-#include <osmocom/msc/osmo_msc.h>
 #include <osmocom/msc/sms_queue.h>
 #include <osmocom/vty/telnet_interface.h>
 #include <osmocom/vty/ports.h>
@@ -61,16 +61,23 @@
 #include <osmocom/msc/smpp.h>
 #include <osmocom/sigtran/osmo_ss7.h>
 #include <osmocom/mgcp_client/mgcp_client.h>
+#include <osmocom/msc/sgs_iface.h>
+#include <osmocom/msc/sgs_server.h>
+#include <osmocom/msc/ran_infra.h>
+#include <osmocom/msc/ran_peer.h>
+#include <osmocom/msc/ran_msg_a.h>
+#include <osmocom/msc/msub.h>
+#include <osmocom/msc/call_leg.h>
+#include <osmocom/msc/msc_ho.h>
+#include <osmocom/msc/gsup_client_mux.h>
+#include <osmocom/msc/e_link.h>
+#include <osmocom/msc/mncc_call.h>
 
 #ifdef BUILD_IU
 #include <osmocom/ranap/iu_client.h>
+#include <asn1c/asn_internal.h>
 #endif
 
-#include <osmocom/msc/msc_ifaces.h>
-#include <osmocom/msc/iucs.h>
-#include <osmocom/msc/iucs_ranap.h>
-#include <osmocom/msc/a_iface.h>
-
 static const char * const osmomsc_copyright =
 	"OsmoMSC - Osmocom Circuit-Switched Core Network implementation\r\n"
 	"Copyright (C) 2016 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>\r\n"
@@ -95,20 +102,15 @@ static struct {
 	const char *config_file;
 	int daemonize;
 	const char *mncc_sock_path;
-	int use_db_counter;
 } msc_cmdline_config = {
-	"sms.db",
-	"osmo-msc.cfg",
-	0,
-	0,
-	1
+	.database_name = "sms.db",
+	.config_file = "osmo-msc.cfg",
 };
 
 /* timer to store statistics */
-#define DB_SYNC_INTERVAL	60, 0
 #define EXPIRE_INTERVAL		10, 0
 
-static struct osmo_timer_list db_sync_timer;
+static int quit = 0;
 
 static void print_usage()
 {
@@ -125,10 +127,9 @@ static void print_help()
 	printf("  -s --disable-color\n");
 	printf("  -l --database db-name      The database to use.\n");
 	printf("  -T --timestamp             Prefix every log line with a timestamp.\n");
-	printf("  -V --version               Print the version of OpenBSC.\n");
+	printf("  -V --version               Print the version of OsmoMSC.\n");
 	printf("  -e --log-level number      Set a global loglevel.\n");
 	printf("  -M --mncc-sock-path PATH   Disable built-in MNCC handler and offer socket.\n");
-	printf("  -C --no-dbcounter          Disable regular syncing of counters to database.\n");
 }
 
 static void handle_options(int argc, char **argv)
@@ -146,7 +147,7 @@ static void handle_options(int argc, char **argv)
 			{"version", 0, 0, 'V' },
 			{"log-level", 1, 0, 'e'},
 			{"mncc-sock-path", 1, 0, 'M'},
-			{"no-dbcounter", 0, 0, 'C'},
+			{"no-dbcounter", 0, 0, 'C'}, /* deprecated */
 			{0, 0, 0, 0}
 		};
 
@@ -185,7 +186,7 @@ static void handle_options(int argc, char **argv)
 			msc_cmdline_config.mncc_sock_path = optarg;
 			break;
 		case 'C':
-			msc_cmdline_config.use_db_counter = 0;
+			fprintf(stderr, "-C is deprecated and does nothing.");
 			break;
 		case 'V':
 			print_version(1);
@@ -197,6 +198,11 @@ static void handle_options(int argc, char **argv)
 			exit(-1);
 		}
 	}
+
+	if (argc > optind) {
+		fprintf(stderr, "Unsupported positional arguments on command line\n");
+		exit(2);
+	}
 }
 
 struct gsm_network *msc_network_alloc(void *ctx,
@@ -214,6 +220,8 @@ struct gsm_network *msc_network_alloc(void *ctx,
 	net->gsup_server_port = MSC_HLR_REMOTE_PORT_DEFAULT;
 
 	mgcp_client_conf_init(&net->mgw.conf);
+	net->mgw.tdefs = g_mgw_tdefs;
+	net->call_waiting = true;
 
 	return net;
 }
@@ -233,10 +241,8 @@ static void signal_handler(int signal)
 	switch (signal) {
 	case SIGINT:
 	case SIGTERM:
-		msc_network_shutdown(msc_network);
-		osmo_signal_dispatch(SS_L_GLOBAL, S_L_GLOBAL_SHUTDOWN, NULL);
-		sleep(3);
-		exit(0);
+		LOGP(DMSC, LOGL_NOTICE, "Terminating due to signal %d\n", signal);
+		quit++;
 		break;
 	case SIGABRT:
 		osmo_generate_backtrace();
@@ -254,19 +260,6 @@ static void signal_handler(int signal)
 	}
 }
 
-/* timer handling */
-static int _db_store_counter(struct osmo_counter *counter, void *data)
-{
-	return db_store_counter(counter);
-}
-
-static void db_sync_timer_cb(void *data)
-{
-	/* store counters to database and re-schedule */
-	osmo_counters_for_each(_db_store_counter, NULL);
-	osmo_timer_schedule(&db_sync_timer, DB_SYNC_INTERVAL);
-}
-
 static int msc_vty_go_parent(struct vty *vty)
 {
 	switch (vty->node) {
@@ -320,27 +313,6 @@ static struct vty_app_info msc_vty_info = {
 	.is_config_node	= msc_vty_is_config_node,
 };
 
-#ifdef BUILD_IU
-static int rcvmsg_iu_cs(struct msgb *msg, struct gprs_ra_id *ra_id, uint16_t *sai)
-{
-	DEBUGP(DIUCS, "got IuCS message %d bytes: %s\n", msg->len, msgb_hexdump(msg));
-	if (ra_id) {
-		DEBUGP(DIUCS, "got IuCS message on %s\n", osmo_rai_name(ra_id));
-	}
-
-	return gsm0408_rcvmsg_iucs(msc_network, msg, ra_id? &ra_id->lac : NULL);
-}
-
-static int rx_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type type,
-		       void *data)
-{
-	DEBUGP(DIUCS, "got IuCS event %u: %s\n", type,
-	       ranap_iu_event_type_str(type));
-
-	return iucs_rx_ranap_event(msc_network, ctx, type, data);
-}
-#endif
-
 #define DEFAULT_M3UA_REMOTE_IP "127.0.0.1"
 #define DEFAULT_PC "0.23.1"
 
@@ -361,43 +333,39 @@ static struct osmo_sccp_instance *sccp_setup(void *ctx, uint32_t cs7_instance,
 	 * Similarly, 'cs7' / 'listen' can specify the local IP address. */
 }
 
-static int ss7_setup(void *ctx)
+static int ss7_setup(void *ctx, struct osmo_sccp_instance **sccp_a, struct osmo_sccp_instance **sccp_iu)
 {
-	uint32_t cs7_instance_a = msc_network->a.cs7_instance;
+	uint32_t i_a = msc_network->a.cs7_instance;
+	uint32_t i_iu = msc_network->iu.cs7_instance;
+
+	const char *name_a = "OsmoMSC-A";
+	const char *name_iu = NULL;
+
 #if BUILD_IU
-	uint32_t cs7_instance_iu = msc_network->iu.cs7_instance;
-
-	if (cs7_instance_a == cs7_instance_iu) {
-		/* Create one single SCCP instance which will be used for both,
-		 * Iu and A at the same time, under the same point-code */
-		LOGP(DMSC, LOGL_NOTICE, "CS7 Instance identifiers: A = Iu = %u\n", cs7_instance_a);
-
-		msc_network->a.sccp = sccp_setup(ctx, cs7_instance_a, "OsmoMSC-A-Iu", DEFAULT_PC);
-		if (!msc_network->a.sccp)
-			return -EINVAL;
-
-		msc_network->iu.sccp = msc_network->a.sccp;
-	} else {
-		/* Create two separate SCCP instances to run A and Iu independently on different
-		 * pointcodes */
-		LOGP(DMSC, LOGL_NOTICE, "CS7 Instance identifiers: A = %u, Iu = %u\n",
-		     cs7_instance_a, cs7_instance_iu);
-
-		msc_network->a.sccp = sccp_setup(ctx, cs7_instance_a, "OsmoMSC-A", DEFAULT_PC);
-		if (!msc_network->a.sccp)
-			return -EINVAL;
-
-		msc_network->iu.sccp = sccp_setup(ctx, cs7_instance_iu, "OsmoMSC-Iu", DEFAULT_PC);
-		if (!msc_network->iu.sccp)
-			return -EINVAL;
-	}
-#else
-	/* No Iu support, just open up an A instance */
-	msc_network->a.sccp = sccp_setup(ctx, cs7_instance_a, "OsmoMSC-A", DEFAULT_PC);
-	if (!msc_network->a.sccp)
-		return -EINVAL;
+	if (i_a == i_iu)
+		name_a = name_iu = "OsmoMSC-A-Iu";
+	else
+		name_iu = "OsmoMSC-Iu";
 #endif
 
+	*sccp_a = sccp_setup(ctx, i_a, name_a, DEFAULT_PC);
+	if (!*sccp_a)
+		return -EINVAL;
+
+	if (!name_iu) {
+		*sccp_iu = NULL;
+		return 0;
+	}
+
+	if (i_a == i_iu) {
+		*sccp_iu = *sccp_a;
+		return 0;
+	}
+
+	*sccp_iu = sccp_setup(ctx, i_iu, name_iu, DEFAULT_PC);
+	if (!*sccp_iu)
+		return -EINVAL;
+
 	return 0;
 }
 
@@ -451,6 +419,7 @@ static const struct log_info_cat msc_default_categories[] = {
 	[DHO] = {
 		.name = "DHO",
 		.description = "Hand-Over",
+		.color = "\033[1;38m",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DDB] = {
@@ -471,29 +440,38 @@ static const struct log_info_cat msc_default_categories[] = {
 	[DSMPP] = {
 		.name = "DSMPP",
 		.description = "SMPP interface for external SMS apps",
-		.enabled = 1, .loglevel = LOGL_DEBUG,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DRANAP] = {
 		.name = "DRANAP",
 		.description = "Radio Access Network Application Part Protocol",
-		.enabled = 1, .loglevel = LOGL_DEBUG,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DVLR] = {
 		.name = "DVLR",
 		.description = "Visitor Location Register",
-		.enabled = 1, .loglevel = LOGL_DEBUG,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DIUCS] = {
 		.name = "DIUCS",
 		.description = "Iu-CS Protocol",
-		.enabled = 1, .loglevel = LOGL_DEBUG,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DBSSAP] = {
 		.name = "DBSSAP",
 		.description = "BSSAP Protocol (A Interface)",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
-
+	[DSGS] = {
+		.name = "DSGS",
+		.description = "SGs Interface (SGsAP)",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
+	[DSS] = {
+		.name = "DSS",
+		.description = "Supplementary Services",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 };
 
 static int filter_fn(const struct log_context *ctx, struct log_target *tar)
@@ -521,43 +499,58 @@ int main(int argc, char **argv)
 {
 	int rc;
 
+	struct osmo_sccp_instance *sccp_a;
+	struct osmo_sccp_instance *sccp_iu;
+
+	/* Track the use of talloc NULL memory contexts */
+	talloc_enable_null_tracking();
+
+	osmo_fsm_set_dealloc_ctx(OTC_SELECT);
+
 	msc_vty_info.copyright	= osmomsc_copyright;
 
 	tall_msc_ctx = talloc_named_const(NULL, 1, "osmo_msc");
 	msc_vty_info.tall_ctx = tall_msc_ctx;
 
 	msgb_talloc_ctx_init(tall_msc_ctx, 0);
+	osmo_signal_talloc_ctx_init(tall_msc_ctx);
 	tall_gsms_ctx = talloc_named_const(tall_msc_ctx, 0, "sms");
 	tall_call_ctx = talloc_named_const(tall_msc_ctx, 0, "gsm_call");
 	tall_trans_ctx = talloc_named_const(tall_msc_ctx, 0, "transaction");
 
 	osmo_init_logging2(tall_msc_ctx, &log_info);
+
+	osmo_fsm_log_timeouts(true);
+	osmo_fsm_log_addr(true);
+
 	osmo_stats_init(tall_msc_ctx);
 
 	/* For --version, vty_init() must be called before handling options */
 	vty_init(&msc_vty_info);
 
-	osmo_ss7_init();
+	OSMO_ASSERT(osmo_ss7_init() == 0);
 	osmo_ss7_vty_init_asp(tall_msc_ctx);
+	osmo_sccp_vty_init();
 
 	/* Parse options */
 	handle_options(argc, argv);
 
-	/* Allocate global gsm_network struct; choose socket/internal MNCC */
-	msc_network = msc_network_alloc(tall_msc_ctx,
-				        msc_cmdline_config.mncc_sock_path?
-						mncc_sock_from_cc
-						: int_mncc_recv);
+	/* Allocate global gsm_network struct.
+	 * At first set the internal MNCC as default, may be changed below according to cfg or cmdline option. */
+	msc_network = msc_network_alloc(tall_msc_ctx, int_mncc_recv);
 	if (!msc_network)
 		return -ENOMEM;
 
+	call_leg_init(msc_network);
+	mncc_call_fsm_init(msc_network);
+
 	if (msc_vlr_alloc(msc_network)) {
 		fprintf(stderr, "Failed to allocate VLR\n");
 		exit(1);
 	}
 
 	ctrl_vty_init(tall_msc_ctx);
-	logging_vty_add_cmds(&log_info);
+	logging_vty_add_cmds();
 	osmo_talloc_vty_add_cmds();
 	msc_vty_init(msc_network);
 
@@ -565,6 +558,7 @@ int main(int argc, char **argv)
 	if (smpp_openbsc_alloc_init(tall_msc_ctx) < 0)
 		return -1;
 #endif
+	sgs_iface_init(tall_msc_ctx, msc_network);
 
 	rc = vty_read_config_file(msc_cmdline_config.config_file, NULL);
 	if (rc < 0) {
@@ -573,10 +567,20 @@ int main(int argc, char **argv)
 		return 1;
 	}
 
-	/* Initialize MNCC socket if appropriate */
+	/* Initialize MNCC socket if appropriate. If the cmdline option -M is present, it overrides the .cfg file
+	 * setting 'msc' / 'mncc external MNCC_SOCKET_PATH'. Note that when -M is given, it "bleeds" back into the vty
+	 * 'write' command and is reflected in the written out 'mncc external' cfg. */
 	if (msc_cmdline_config.mncc_sock_path) {
+		LOGP(DMNCC, LOGL_NOTICE,
+		     "MNCC socket path is configured from commandline argument -M."
+		     " This affects a written-back config file. Instead consider using the config file directly"
+		     " ('msc' / 'mncc external MNCC_SOCKET_PATH').\n");
+		gsm_network_set_mncc_sock_path(msc_network, msc_cmdline_config.mncc_sock_path);
+	}
+	if (msc_network->mncc_sock_path) {
+		msc_network->mncc_recv = mncc_sock_from_cc;
 		rc = mncc_sock_init(msc_network,
-				    msc_cmdline_config.mncc_sock_path);
+				    msc_network->mncc_sock_path);
 		if (rc) {
 			fprintf(stderr, "MNCC socket initialization failed. exiting.\n");
 			exit(1);
@@ -631,24 +635,23 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 		return 4;
 	}
 
-	osmo_fsm_log_addr(true);
+	if (msc_gsup_client_start(msc_network)) {
+		fprintf(stderr, "Failed to start GSUP client\n");
+		exit(1);
+	}
+
+	msc_a_i_t_gsup_init(msc_network);
+
 	if (msc_vlr_start(msc_network)) {
 		fprintf(stderr, "Failed to start VLR\n");
 		exit(1);
 	}
 
-	msc_subscr_conn_init();
-
 	if (db_prepare()) {
 		printf("DB: Failed to prepare database.\n");
 		return 5;
 	}
 
-	/* setup the timer */
-	osmo_timer_setup(&db_sync_timer, db_sync_timer_cb, NULL);
-	if (msc_cmdline_config.use_db_counter)
-		osmo_timer_schedule(&db_sync_timer, DB_SYNC_INTERVAL);
-
 	signal(SIGINT, &signal_handler);
 	signal(SIGTERM, &signal_handler);
 	signal(SIGABRT, &signal_handler);
@@ -668,18 +671,50 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 		return 7;
 	}
 
-	if (ss7_setup(tall_msc_ctx)) {
+	if (ss7_setup(tall_msc_ctx, &sccp_a, &sccp_iu)) {
 		printf("Setting up SCCP client failed.\n");
 		return 8;
 	}
 
+	if (sgs_server_open(g_sgs)) {
+		printf("Starting SGs server failed\n");
+		return 9;
+	}
+
+	msc_network->a.sri = sccp_ran_init(msc_network, sccp_a, OSMO_SCCP_SSN_BSSAP,
+					   "OsmoMSC-A", &msc_ran_infra[OSMO_RAT_GERAN_A],
+					   msc_network);
+	if (!msc_network->a.sri) {
+		printf("Setting up A receiver failed\n");
+		return 10;
+	}
+	LOGP(DMSC, LOGL_NOTICE, "A-interface: SCCP user %s, cs7-instance %u (%s)\n",
+	     osmo_sccp_user_name(msc_network->a.sri->scu),
+	     osmo_sccp_get_ss7(msc_network->a.sri->sccp)->cfg.id,
+	     osmo_sccp_get_ss7(msc_network->a.sri->sccp)->cfg.name);
+
 #ifdef BUILD_IU
-	/* Set up IuCS */
-	ranap_iu_init(tall_msc_ctx, DRANAP, "OsmoMSC-IuCS", msc_network->iu.sccp, rcvmsg_iu_cs, rx_iu_event);
+	talloc_asn1_ctx = talloc_named_const(tall_msc_ctx, 0, "asn1");
+
+	msc_network->iu.sri = sccp_ran_init(msc_network, sccp_iu, OSMO_SCCP_SSN_RANAP,
+					   "OsmoMSC-IuCS", &msc_ran_infra[OSMO_RAT_UTRAN_IU],
+					   msc_network);
+	if (!msc_network->iu.sri) {
+		printf("Setting up IuCS receiver failed\n");
+		return 11;
+	}
+
+	/* Compatibility with legacy osmo-hnbgw that was unable to properly handle RESET messages. */
+	msc_network->iu.sri->ignore_missing_reset = true;
+
+	LOGP(DMSC, LOGL_NOTICE, "Iu-interface: SCCP user %s, cs7-instance %u (%s)\n",
+	     osmo_sccp_user_name(msc_network->iu.sri->scu),
+	     osmo_sccp_get_ss7(msc_network->iu.sri->sccp)->cfg.id,
+	     osmo_sccp_get_ss7(msc_network->iu.sri->sccp)->cfg.name);
 #endif
 
-	/* Set up A interface */
-	a_init(msc_network->a.sccp, msc_network);
+	/* Init RRLP handlers */
+	msc_rrlp_init();
 
 	if (msc_cmdline_config.daemonize) {
 		rc = osmo_daemonize();
@@ -689,8 +724,32 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 		}
 	}
 
-	while (1) {
+	while (!quit) {
 		log_reset_context();
-		osmo_select_main(0);
+		osmo_select_main_ctx(0);
 	}
+
+	msc_network_shutdown(msc_network);
+	osmo_signal_dispatch(SS_L_GLOBAL, S_L_GLOBAL_SHUTDOWN, NULL);
+	sleep(3);
+
+	log_fini();
+
+	/**
+	 * Report the heap state of root context, then free,
+	 * so both ASAN and Valgrind are happy...
+	 */
+	talloc_report_full(tall_msc_ctx, stderr);
+	talloc_free(tall_msc_ctx);
+
+	/* FIXME: VTY code still uses NULL-context */
+	talloc_free(tall_vty_ctx);
+
+	/**
+	 * Report the heap state of NULL context, then free,
+	 * so both ASAN and Valgrind are happy...
+	 */
+	talloc_report_full(NULL, stderr);
+	talloc_disable_null_tracking();
+	return 0;
 }
