@@ -25,31 +25,35 @@
 
 #include <osmocom/msc/gsm_04_08.h>
 
-static void mncc_sends_to_cc(uint32_t msg_type, struct gsm_mncc *mncc)
-{
-	mncc->msg_type = msg_type;
-	mncc_tx_to_cc(net, msg_type, mncc);
-}
+#define mncc_sends_to_cc(MSG_TYPE, MNCC) do { \
+		(MNCC)->msg_type = MSG_TYPE; \
+		log("MSC <-- MNCC: callref 0x%x: %s\n%s", (MNCC)->callref, \
+		    get_mncc_name((MNCC)->msg_type), \
+		    (MNCC)->sdp); \
+		mncc_tx_to_cc(net, MNCC); \
+	} while(0)
 
+/*
 static void on_call_release_mncc_sends_to_cc(uint32_t msg_type, struct gsm_mncc *mncc)
 {
 	mncc->msg_type = msg_type;
 	on_call_release_mncc_sends_to_cc_data = mncc;
 }
+*/
 
 #define IMSI "901700000010650"
 
-static void standard_lu()
+static void lu_utran_tmsi()
 {
 	struct vlr_subscr *vsub;
 
 	net->authentication_required = true;
 	net->vlr->cfg.assign_tmsi = true;
-	rx_from_ran = RAN_UTRAN_IU;
+	rx_from_ran = OSMO_RAT_UTRAN_IU;
 
 	btw("Location Update request causes a GSUP Send Auth Info request to HLR");
 	lu_result_sent = RES_NONE;
-	gsup_expect_tx("080108" "09710000000156f0");
+	gsup_expect_tx("080108" "09710000000156f0" CN_DOMAIN VLR_TO_HLR);
 	ms_sends_msg("0508" /* MM LU */
 		     "7" /* ciph key seq: no key available */
 		     "0" /* LU type: normal */
@@ -110,7 +114,8 @@ static void standard_lu()
 			"2310" "65af0527fda95b0dc5ae4aa515cdf32f"
 			"2410" "537c3b35a3b13b08d08eeb28098f45cc"
 			"2510" "4bf4e564f75300009bc796706bc65744"
-			"2708" "0edb0eadbea94ac2",
+			"2708" "0edb0eadbea94ac2"
+		HLR_TO_VLR,
 		NULL);
 	VERBOSE_ASSERT(auth_request_sent, == true, "%d");
 	VERBOSE_ASSERT(lu_result_sent, == RES_NONE, "%d");
@@ -122,18 +127,18 @@ static void standard_lu()
 	VERBOSE_ASSERT(lu_result_sent, == RES_NONE, "%d");
 
 	btw("MS sends SecurityModeControl acceptance, VLR accepts and sends GSUP LU Req to HLR");
-	gsup_expect_tx("04010809710000000156f0");
+	gsup_expect_tx("04010809710000000156f0" CN_DOMAIN VLR_TO_HLR);
 	ms_sends_security_mode_complete();
 	VERBOSE_ASSERT(gsup_tx_confirmed, == true, "%d");
 	VERBOSE_ASSERT(lu_result_sent, == RES_NONE, "%d");
 
 	btw("HLR sends _INSERT_DATA_REQUEST, VLR responds with _INSERT_DATA_RESULT");
-	gsup_rx("10010809710000000156f00804032443f2",
-		"12010809710000000156f0");
+	gsup_rx("10010809710000000156f00804032443f2" HLR_TO_VLR,
+		"12010809710000000156f0" VLR_TO_HLR);
 	VERBOSE_ASSERT(lu_result_sent, == RES_NONE, "%d");
 
 	btw("HLR also sends GSUP _UPDATE_LOCATION_RESULT");
-	gsup_rx("06010809710000000156f0", NULL);
+	gsup_rx("06010809710000000156f0" HLR_TO_VLR, NULL);
 
 	VERBOSE_ASSERT(lu_result_sent, == RES_ACCEPT, "%d");
 
@@ -148,14 +153,14 @@ static void standard_lu()
 	VERBOSE_ASSERT(iu_release_sent, == true, "%d"); \
 
 	btw("LU was successful, and the conn has already been closed");
-	rnc_sends_release_complete();
+	ran_sends_clear_complete();
 	EXPECT_CONN_COUNT(0);
 
-	vsub = vlr_subscr_find_by_imsi(net->vlr, IMSI);
+	vsub = vlr_subscr_find_by_imsi(net->vlr, IMSI, __func__);
 	VERBOSE_ASSERT(vsub != NULL, == true, "%d");
 	VERBOSE_ASSERT(strcmp(vsub->imsi, IMSI), == 0, "%d");
-	VERBOSE_ASSERT(vsub->lac, == 23, "%u");
-	vlr_subscr_put(vsub);
+	VAL_ASSERT("LAC", vsub->cgi.lai.lac, == 23, "%u");
+	vlr_subscr_put(vsub, __func__);
 }
 
 static void test_call_mo()
@@ -168,19 +173,16 @@ static void test_call_mo()
 
 	fake_time_start();
 
-	standard_lu();
+	lu_utran_tmsi();
 
 	BTW("after a while, a new conn sends a CM Service Request. VLR responds with Auth Req, 2nd auth vector");
 	auth_request_sent = false;
 	auth_request_expect_rand = "c187a53a5e6b9d573cac7c74451fd46d";
 	auth_request_expect_autn = "1843a645b98d00005b2d666af46c45d9";
 	cm_service_result_sent = RES_NONE;
-	ms_sends_msg("052478"
+	ms_sends_msg("052471"
 		     "03575886" /* classmark 2 */
 		     "089910070000106005" /* IMSI */);
-	OSMO_ASSERT(g_conn);
-	OSMO_ASSERT(g_conn->fi);
-	OSMO_ASSERT(g_conn->vsub);
 	VERBOSE_ASSERT(cm_service_result_sent, == RES_NONE, "%d");
 	VERBOSE_ASSERT(auth_request_sent, == true, "%d");
 
@@ -213,6 +215,25 @@ static void test_call_mo()
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	mncc.callref = cc_to_mncc_tx_got_callref;
 
+	btw("MNCC replies with MNCC_RTP_CREATE, causing MGW endpoint CRCX to RAN");
+	expect_crcx(RTP_TO_RAN);
+	mncc_sends_to_cc(MNCC_RTP_CREATE, &mncc);
+	OSMO_ASSERT(got_crcx);
+
+	btw("MGW acknowledges the CRCX, triggering Assignment");
+	expect_iu_rab_assignment();
+	crcx_ok(RTP_TO_RAN);
+	OSMO_ASSERT(iu_rab_assignment_sent);
+
+	btw("Assignment succeeds, triggering CRCX to CN");
+	expect_crcx(RTP_TO_CN);
+	ms_sends_assignment_complete(CODEC_AMR_8000_1);
+	OSMO_ASSERT(got_crcx);
+
+	btw("CN RTP address is available, trigger MNCC_RTP_CREATE");
+	cc_to_mncc_expect_tx("", MNCC_RTP_CREATE);
+	crcx_ok(RTP_TO_CN);
+
 	btw("MNCC says that's fine");
 	dtap_expect_tx("8302" /* CC: Call Proceeding */);
 	mncc_sends_to_cc(MNCC_CALL_PROC_REQ, &mncc);
@@ -253,7 +274,7 @@ static void test_call_mo()
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	OSMO_ASSERT(iu_release_sent);
 
-	rnc_sends_release_complete();
+	ran_sends_clear_complete();
 	EXPECT_CONN_COUNT(0);
 	clear_vlr();
 	comment_end();
@@ -270,16 +291,15 @@ static void test_call_mt()
 
 	fake_time_start();
 
-	standard_lu();
+	lu_utran_tmsi();
 
 	BTW("after a while, MNCC asks us to setup a call, causing Paging");
-	
+
 	paging_expect_imsi(IMSI);
 	paging_sent = false;
 	mncc_sends_to_cc(MNCC_SETUP_REQ, &mncc);
 
 	VERBOSE_ASSERT(paging_sent, == true, "%d");
-	VERBOSE_ASSERT(paging_stopped, == false, "%d");
 
 	btw("MS replies with Paging Response, and VLR sends Auth Request");
 	auth_request_sent = false;
@@ -298,13 +318,33 @@ static void test_call_mt()
 	btw("MS sends SecurityModeControl acceptance, VLR accepts, sends CC Setup");
 	dtap_expect_tx("0305" /* CC: Setup */);
 	ms_sends_security_mode_complete();
-	VERBOSE_ASSERT(paging_stopped, == true, "%d");
 
+	btw("MS confirms call, we create a RAN-side RTP and forward MNCC_CALL_CONF_IND");
+	expect_crcx(RTP_TO_RAN);
 	cc_to_mncc_expect_tx(IMSI, MNCC_CALL_CONF_IND);
 	ms_sends_msg("8348" /* CC: Call Confirmed */
 		     "0406600402000581" /* Bearer Capability */
 		     "15020100" /* Call Control Capabilities */
 		     "40080402600400021f00" /* Supported Codec List */);
+	OSMO_ASSERT(got_crcx);
+	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
+
+	btw("MNCC sends MNCC_RTP_CREATE");
+	mncc_sends_to_cc(MNCC_RTP_CREATE, &mncc);
+
+	btw("MGW acknowledges the CRCX to RAN, triggering Assignment");
+	expect_iu_rab_assignment();
+	crcx_ok(RTP_TO_RAN);
+	OSMO_ASSERT(iu_rab_assignment_sent);
+
+	btw("Assignment completes, triggering CRCX to CN");
+	expect_crcx(RTP_TO_CN);
+	ms_sends_assignment_complete(CODEC_AMR_8000_1);
+	OSMO_ASSERT(got_crcx);
+
+	btw("When the CN side RTP address is known, send MNCC_RTP_CREATE");
+	cc_to_mncc_expect_tx("", MNCC_RTP_CREATE);
+	crcx_ok(RTP_TO_CN);
 
 	fake_time_passes(1, 23);
 
@@ -337,7 +377,7 @@ static void test_call_mt()
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	OSMO_ASSERT(iu_release_sent);
 
-	rnc_sends_release_complete();
+	ran_sends_clear_complete();
 	EXPECT_CONN_COUNT(0);
 	clear_vlr();
 	comment_end();
@@ -354,16 +394,15 @@ static void test_call_mt2()
 
 	fake_time_start();
 
-	standard_lu();
+	lu_utran_tmsi();
 
 	BTW("after a while, MNCC asks us to setup a call, causing Paging");
-	
+
 	paging_expect_imsi(IMSI);
 	paging_sent = false;
 	mncc_sends_to_cc(MNCC_SETUP_REQ, &mncc);
 
 	VERBOSE_ASSERT(paging_sent, == true, "%d");
-	VERBOSE_ASSERT(paging_stopped, == false, "%d");
 
 	btw("MS replies with Paging Response, and VLR sends Auth Request");
 	auth_request_sent = false;
@@ -382,15 +421,33 @@ static void test_call_mt2()
 	btw("MS sends SecurityModeControl acceptance, VLR accepts, sends CC Setup");
 	dtap_expect_tx("0305" /* CC: Setup */);
 	ms_sends_security_mode_complete();
-	VERBOSE_ASSERT(paging_stopped, == true, "%d");
 
+	btw("MS confirms call, we create a RAN-side RTP and forward MNCC_CALL_CONF_IND");
+	expect_crcx(RTP_TO_RAN);
 	cc_to_mncc_expect_tx(IMSI, MNCC_CALL_CONF_IND);
 	ms_sends_msg("8348" /* CC: Call Confirmed */
 		     "0406600402000581" /* Bearer Capability */
 		     "15020100" /* Call Control Capabilities */
 		     "40080402600400021f00" /* Supported Codec List */);
+	OSMO_ASSERT(got_crcx);
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 
+	btw("MNCC sends MNCC_RTP_CREATE");
+	mncc_sends_to_cc(MNCC_RTP_CREATE, &mncc);
+
+	btw("MGW acknowledges the CRCX to RAN, triggering Assignment");
+	expect_iu_rab_assignment();
+	crcx_ok(RTP_TO_RAN);
+	OSMO_ASSERT(iu_rab_assignment_sent);
+
+	btw("Assignment completes, triggering CRCX to CN");
+	expect_crcx(RTP_TO_CN);
+	ms_sends_assignment_complete(CODEC_AMR_8000_1);
+	OSMO_ASSERT(got_crcx);
+
+	btw("When the CN side RTP address is known, send MNCC_RTP_CREATE");
+	cc_to_mncc_expect_tx("", MNCC_RTP_CREATE);
+	crcx_ok(RTP_TO_CN);
 	fake_time_passes(1, 23);
 
 	cc_to_mncc_expect_tx("", MNCC_ALERT_IND);
@@ -400,15 +457,19 @@ static void test_call_mt2()
 	fake_time_passes(15, 23);
 
 	btw("The call failed, the BSC sends a BSSMAP Clear Request");
-	on_call_release_mncc_sends_to_cc(MNCC_REL_REQ, &mncc);
+	/* FIXME: in this scenario, we send an MNCC_REL_CNF even though MNCC never asked us to MNCC_REL_REQ.  Legacy
+	 * behavior did get to both MNCC_REL_IND, then an MNCC_REL_REQ from MNCC as well as a final MNCC_REL_CNF, but
+	 * this only worked synchronously, i.e. only with internal MNCC. Instead of mimicking that, we need a proper
+	 * async solution that also works with a PBX. */
 	cc_to_mncc_expect_tx("", MNCC_REL_CNF);
-	dtap_expect_tx("032d"); /* CC: Release */
+	dtap_expect_tx("032d0802e1af"); /* CC: Release */
 	expect_iu_release();
-	msc_clear_request(g_conn, 0);
+	msc_a_release_cn(msub_msc_a(g_msub));
+	OSMO_ASSERT(dtap_tx_confirmed);
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	OSMO_ASSERT(iu_release_sent);
 
-	rnc_sends_release_complete();
+	ran_sends_clear_complete();
 	EXPECT_CONN_COUNT(0);
 
 	/* Make sure a pending release timer doesn't fire later to access freed data */
@@ -428,19 +489,16 @@ static void test_call_mo_to_unknown()
 
 	fake_time_start();
 
-	standard_lu();
+	lu_utran_tmsi();
 
 	BTW("after a while, a new conn sends a CM Service Request. VLR responds with Auth Req, 2nd auth vector");
 	auth_request_sent = false;
 	auth_request_expect_rand = "c187a53a5e6b9d573cac7c74451fd46d";
 	auth_request_expect_autn = "1843a645b98d00005b2d666af46c45d9";
 	cm_service_result_sent = RES_NONE;
-	ms_sends_msg("052478"
+	ms_sends_msg("052471"
 		     "03575886" /* classmark 2 */
 		     "089910070000106005" /* IMSI */);
-	OSMO_ASSERT(g_conn);
-	OSMO_ASSERT(g_conn->fi);
-	OSMO_ASSERT(g_conn->vsub);
 	VERBOSE_ASSERT(cm_service_result_sent, == RES_NONE, "%d");
 	VERBOSE_ASSERT(auth_request_sent, == true, "%d");
 
@@ -473,6 +531,25 @@ static void test_call_mo_to_unknown()
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	mncc.callref = cc_to_mncc_tx_got_callref;
 
+	btw("MNCC replies with MNCC_RTP_CREATE, causing MGW endpoint CRCX to RAN");
+	expect_crcx(RTP_TO_RAN);
+	mncc_sends_to_cc(MNCC_RTP_CREATE, &mncc);
+	OSMO_ASSERT(got_crcx);
+
+	btw("MGW acknowledges the CRCX, triggering Assignment");
+	expect_iu_rab_assignment();
+	crcx_ok(RTP_TO_RAN);
+	OSMO_ASSERT(iu_rab_assignment_sent);
+
+	btw("Assignment succeeds, triggering CRCX to CN");
+	expect_crcx(RTP_TO_CN);
+	ms_sends_assignment_complete(CODEC_AMR_8000_1);
+	OSMO_ASSERT(got_crcx);
+
+	btw("CN RTP address is available, trigger MNCC_RTP_CREATE");
+	cc_to_mncc_expect_tx("", MNCC_RTP_CREATE);
+	crcx_ok(RTP_TO_CN);
+
 	btw("MNCC says that's fine");
 	dtap_expect_tx("8302" /* CC: Call Proceeding */);
 	mncc_sends_to_cc(MNCC_CALL_PROC_REQ, &mncc);
@@ -492,7 +569,7 @@ static void test_call_mo_to_unknown()
 	OSMO_ASSERT(iu_release_sent);
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 
-	rnc_sends_release_complete();
+	ran_sends_clear_complete();
 	EXPECT_CONN_COUNT(0);
 	clear_vlr();
 	comment_end();
@@ -508,19 +585,16 @@ static void test_call_mo_to_unknown_timeout()
 
 	fake_time_start();
 
-	standard_lu();
+	lu_utran_tmsi();
 
 	BTW("after a while, a new conn sends a CM Service Request. VLR responds with Auth Req, 2nd auth vector");
 	auth_request_sent = false;
 	auth_request_expect_rand = "c187a53a5e6b9d573cac7c74451fd46d";
 	auth_request_expect_autn = "1843a645b98d00005b2d666af46c45d9";
 	cm_service_result_sent = RES_NONE;
-	ms_sends_msg("052478"
+	ms_sends_msg("052471"
 		     "03575886" /* classmark 2 */
 		     "089910070000106005" /* IMSI */);
-	OSMO_ASSERT(g_conn);
-	OSMO_ASSERT(g_conn->fi);
-	OSMO_ASSERT(g_conn->vsub);
 	VERBOSE_ASSERT(cm_service_result_sent, == RES_NONE, "%d");
 	VERBOSE_ASSERT(auth_request_sent, == true, "%d");
 
@@ -553,6 +627,25 @@ static void test_call_mo_to_unknown_timeout()
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	mncc.callref = cc_to_mncc_tx_got_callref;
 
+	btw("MNCC replies with MNCC_RTP_CREATE, causing MGW endpoint CRCX to RAN");
+	expect_crcx(RTP_TO_RAN);
+	mncc_sends_to_cc(MNCC_RTP_CREATE, &mncc);
+	OSMO_ASSERT(got_crcx);
+
+	btw("MGW acknowledges the CRCX, triggering Assignment");
+	expect_iu_rab_assignment();
+	crcx_ok(RTP_TO_RAN);
+	OSMO_ASSERT(iu_rab_assignment_sent);
+
+	btw("Assignment succeeds, triggering CRCX to CN");
+	expect_crcx(RTP_TO_CN);
+	ms_sends_assignment_complete(CODEC_AMR_8000_1);
+	OSMO_ASSERT(got_crcx);
+
+	btw("CN RTP address is available, trigger MNCC_RTP_CREATE");
+	cc_to_mncc_expect_tx("", MNCC_RTP_CREATE);
+	crcx_ok(RTP_TO_CN);
+
 	btw("MNCC says that's fine");
 	dtap_expect_tx("8302" /* CC: Call Proceeding */);
 	mncc_sends_to_cc(MNCC_CALL_PROC_REQ, &mncc);
@@ -573,7 +666,7 @@ static void test_call_mo_to_unknown_timeout()
 	OSMO_ASSERT(cc_to_mncc_tx_confirmed);
 	OSMO_ASSERT(iu_release_sent);
 
-	rnc_sends_release_complete();
+	ran_sends_clear_complete();
 	EXPECT_CONN_COUNT(0);
 	clear_vlr();
 	comment_end();
