@@ -27,14 +27,20 @@
 #include <stdio.h>
 
 #include <osmocom/msc/gsm_data.h>
-#include <osmocom/msc/osmo_msc.h>
 #include <osmocom/msc/vlr.h>
+#include <osmocom/msc/msub.h>
+#include <osmocom/msc/msc_a.h>
 #include <osmocom/msc/mncc.h>
+#include <osmocom/msc/call_leg.h>
+#include <osmocom/msc/rtp_stream.h>
 
 extern bool _log_lines;
+#define LOG_COLOR "\033[1;33m"
+#define LOG_COLOR_OFF "\033[0;m"
+
 #define _log(fmt, args...) do { \
 		if (_log_lines) \
-			fprintf(stderr, " %4d:%s: " fmt "\n", \
+			fprintf(stderr, LOG_COLOR " %4d:%s: " fmt LOG_COLOR_OFF "\n", \
 				__LINE__, __FILE__, ## args ); \
 		else \
 			fprintf(stderr, fmt "\n", ## args ); \
@@ -42,19 +48,18 @@ extern bool _log_lines;
 
 /* btw means "by the way", the test tells the log what's happening.
  * BTW() marks a larger section, btw() is the usual logging. */
-#define BTW(fmt, args...) _log("---\n- " fmt, ## args )
+#define BTW(fmt, args...) _log("\n\n- " fmt, ## args )
 #define btw(fmt, args...) _log("- " fmt, ## args )
 #define log(fmt, args...) _log("  " fmt, ## args )
 
 #define comment_start() fprintf(stderr, "===== %s\n", __func__);
 #define comment_end() fprintf(stderr, "===== %s: SUCCESS\n\n", __func__);
 
-extern struct gsm_subscriber_connection *g_conn;
+extern struct msub *g_msub;
 extern struct gsm_network *net;
-extern struct gsm_bts *the_bts;
 extern void *msgb_ctx;
 
-extern enum ran_type rx_from_ran;
+extern enum osmo_rat_type rx_from_ran;
 
 extern const char *gsup_tx_expected;
 extern bool gsup_tx_confirmed;
@@ -100,7 +105,6 @@ static inline void expect_security_mode_ctrl(const char *ck, const char *ik)
 }
 
 extern bool paging_sent;
-extern bool paging_stopped;
 
 extern bool iu_release_expected;
 extern bool iu_release_sent;
@@ -126,13 +130,13 @@ static inline void expect_bssap_clear()
 	bssap_clear_sent = false;
 }
 
-static inline void expect_release_clear(enum ran_type via_ran)
+static inline void expect_release_clear(enum osmo_rat_type via_ran)
 {
 	switch (via_ran) {
-	case RAN_GERAN_A:
+	case OSMO_RAT_GERAN_A:
 		expect_bssap_clear();
 		return;
-	case RAN_UTRAN_IU:
+	case OSMO_RAT_UTRAN_IU:
 		expect_iu_release();
 		return;
 	default:
@@ -141,6 +145,23 @@ static inline void expect_release_clear(enum ran_type via_ran)
 	}
 }
 
+extern bool bssap_assignment_expected;
+extern bool bssap_assignment_sent;
+extern bool iu_rab_assignment_expected;
+extern bool iu_rab_assignment_sent;
+
+static inline void expect_bssap_assignment()
+{
+	bssap_assignment_expected = true;
+	bssap_assignment_sent = false;
+}
+
+static inline void expect_iu_rab_assignment()
+{
+	iu_rab_assignment_expected = true;
+	iu_rab_assignment_sent = false;
+}
+
 struct msc_vlr_test_cmdline_opts {
 	bool verbose;
 	int run_test_nr;
@@ -152,7 +173,8 @@ extern msc_vlr_test_func_t msc_vlr_tests[];
 struct msgb *msgb_from_hex(const char *label, uint16_t size, const char *hex);
 
 void clear_vlr();
-bool conn_exists(struct gsm_subscriber_connection *conn);
+bool conn_exists(const struct msub *msub);
+void conn_conclude_cm_service_req(struct msub *msub, const char *cm_service_use);
 
 void dtap_expect_tx(const char *hex);
 void dtap_expect_tx_ussd(char *ussd_text);
@@ -160,33 +182,38 @@ void paging_expect_imsi(const char *imsi);
 void paging_expect_tmsi(uint32_t tmsi);
 
 void ms_sends_msg(const char *hex);
+void ms_sends_classmark_update(const struct osmo_gsm48_classmark *classmark);
+void ms_sends_ciphering_mode_complete(const char *inner_nas_msg);
 void ms_sends_security_mode_complete();
+void ms_sends_assignment_complete(enum mgcp_codecs assigned_codec);
 void gsup_rx(const char *rx_hex, const char *expect_tx_hex);
 void send_sms(struct vlr_subscr *receiver,
 	      struct vlr_subscr *sender,
 	      char *str);
+void crcx_ok(enum rtp_direction dir);
 
-void bss_sends_clear_complete();
-void rnc_sends_release_complete();
+void ran_sends_clear_complete();
 
 void thwart_rx_non_initial_requests();
 
 #define EXPECT_ACCEPTED(expect_accepted) do { \
-		if (g_conn) \
-			OSMO_ASSERT(conn_exists(g_conn)); \
-		bool accepted = msc_subscr_conn_is_accepted(g_conn); \
-		fprintf(stderr, "msc_subscr_conn_is_accepted() == %s\n", \
+		if (g_msub) \
+			OSMO_ASSERT(conn_exists(g_msub)); \
+		bool accepted = msc_a_is_accepted(msub_msc_a(g_msub)); \
+		fprintf(stderr, "msc_a_is_accepted() == %s\n", \
 			accepted ? "true" : "false"); \
 		OSMO_ASSERT(accepted == expect_accepted); \
 	} while (false)
 
-#define VERBOSE_ASSERT(val, expect_op, fmt) \
+#define VAL_ASSERT(desc, val, expect_op, fmt)	\
 	do { \
-		log(#val " == " fmt, (val)); \
+		log(desc " == " fmt, (val)); \
 		OSMO_ASSERT((val) expect_op); \
-	} while (0);
+	} while (0)
 
-#define EXPECT_CONN_COUNT(N) VERBOSE_ASSERT(llist_count(&net->subscr_conns), == N, "%d")
+#define VERBOSE_ASSERT(val, expect_op, fmt) VAL_ASSERT(#val, val, expect_op, fmt)
+
+#define EXPECT_CONN_COUNT(N) VERBOSE_ASSERT(llist_count(&msub_list), == N, "%d")
 
 #define gsup_expect_tx(hex) do \
 { \
@@ -218,6 +245,7 @@ void fake_time_start();
 { \
 	struct timeval diff; \
 	osmo_gettimeofday_override_add(secs, usecs); \
+	osmo_clock_override_add(CLOCK_MONOTONIC, secs, usecs * 1000); \
 	timersub(&osmo_gettimeofday_override_time, &fake_time_start_time, &diff); \
 	btw("Total time passed: %d.%06d s", \
 	    (int)diff.tv_sec, (int)diff.tv_usec); \
@@ -229,10 +257,10 @@ extern const struct timeval fake_time_start_time;
 
 #define ASSERT_RELEASE_CLEAR(via_ran) \
 	switch (via_ran) { \
-	case RAN_GERAN_A: \
+	case OSMO_RAT_GERAN_A: \
 		VERBOSE_ASSERT(bssap_clear_sent, == true, "%d"); \
 		break; \
-	case RAN_UTRAN_IU: \
+	case OSMO_RAT_UTRAN_IU: \
 		VERBOSE_ASSERT(iu_release_sent, == true, "%d"); \
 		break; \
 	default: \
@@ -240,17 +268,11 @@ extern const struct timeval fake_time_start_time;
 		break; \
 	}
 
-static inline void bss_rnc_sends_release_clear_complete(enum ran_type via_ran)
-{
-	switch (via_ran) {
-	case RAN_GERAN_A:
-		bss_sends_clear_complete();
-		return;
-	case RAN_UTRAN_IU:
-		rnc_sends_release_complete();
-		return;
-	default:
-		OSMO_ASSERT(false);
-		break;
-	}
-}
+#define HLR_TO_VLR "0a0101"
+#define VLR_TO_HLR "0a0101"
+#define CN_DOMAIN "280102"
+#define EUSE_TO_MSC_USSD "0a0103"
+#define MSC_USSD_TO_EUSE "0a0103"
+
+extern bool got_crcx;
+void expect_crcx(enum rtp_direction towards);
